import json
from copy import deepcopy
from unittest.mock import patch

from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.test import TestCase
from django.utils.timezone import now, timedelta
from freezegun import freeze_time
from swapper import load_model

from openwisp_controller.connection.tasks import update_config
from openwisp_controller.connection.tests.base import CreateConnectionsMixin
from openwisp_utils.tests import catch_signal

from ...db import timeseries_db
from .. import settings as app_settings
from ..signals import health_status_changed
from ..tasks import delete_wifi_clients_and_sessions, trigger_device_checks
from ..utils import get_device_cache_key
from . import DeviceMonitoringTestCase, TestWifiClientSessionMixin

DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
DeviceData = load_model('device_monitoring', 'DeviceData')
WifiClient = load_model('device_monitoring', 'WifiClient')
WifiSession = load_model('device_monitoring', 'WifiSession')


class BaseTestCase(DeviceMonitoringTestCase):
    _sample_data = {
        "type": "DeviceMonitoring",
        "general": {
            "hostname": "10-BA-00-FF-E1-9B",
            "local_time": 1519645078,
            "uptime": 1519645078,
        },
        "resources": {
            "load": [10144, 15456, 6624],
            "memory": {
                "buffered": 3334144,
                'cached': 6774784,
                "free": 79429632,
                "shared": 102400,
                "total": 128430080,
            },
            'disk': [
                {
                    'used_bytes': 18792,
                    'available_bytes': 233984,
                    'filesystem': '/dev/root',
                    'mount_point': '/',
                    'used_percent': 7,
                    'size_bytes': 258016,
                }
            ],
            "swap": {"free": 0, "total": 0},
        },
        "interfaces": [
            {
                "name": "wlan0",
                "statistics": {
                    "collisions": 0,
                    "multicast": 0,
                    "rx_bytes": 0,
                    "rx_compressed": 0,
                    "rx_crc_errors": 0,
                    "rx_dropped": 0,
                    "rx_errors": 0,
                    "rx_fifo_errors": 0,
                    "rx_frame_errors": 0,
                    "rx_length_errors": 0,
                    "rx_missed_errors": 0,
                    "rx_over_errors": 0,
                    "rx_packets": 0,
                    "tx_aborted_errors": 0,
                    "tx_bytes": 864,
                    "tx_carrier_errors": 0,
                    "tx_compressed": 0,
                    "tx_dropped": 0,
                    "tx_errors": 0,
                    "tx_fifo_errors": 0,
                    "tx_heartbeat_errors": 0,
                    "tx_packets": 7,
                    "tx_window_errors": 0,
                },
                "wireless": {
                    "channel": 1,
                    "country": "US",
                    "mode": "access_point",
                    "tx_power": 6,
                    "frequency": 5180,
                    "mode": "access_point",
                    "noise": -103,
                    "signal": -56,
                    "ssid": "wifi-test",
                    "htmode": "HT20",
                    "clients": [
                        {
                            "aid": 1,
                            "assoc": True,
                            "auth": True,
                            "authorized": True,
                            "ht": True,
                            "mac": "44:D1:FA:4B:00:00",
                            "mfp": False,
                            "preauth": False,
                            "rrm": [0, 0, 0, 0, 0],
                            "signature": "test_signature",
                            "vht": True,
                            "wds": False,
                            "wmm": True,
                            "wps": False,
                        }
                    ],
                },
            },
            {
                "name": "wlan1",
                "statistics": {
                    "collisions": 0,
                    "multicast": 0,
                    "rx_bytes": 0,
                    "rx_compressed": 0,
                    "rx_crc_errors": 0,
                    "rx_dropped": 0,
                    "rx_errors": 0,
                    "rx_fifo_errors": 0,
                    "rx_frame_errors": 0,
                    "rx_length_errors": 0,
                    "rx_missed_errors": 0,
                    "rx_over_errors": 0,
                    "rx_packets": 0,
                    "tx_aborted_errors": 0,
                    "tx_bytes": 136840,
                    "tx_carrier_errors": 0,
                    "tx_compressed": 0,
                    "tx_dropped": 0,
                    "tx_errors": 0,
                    "tx_fifo_errors": 0,
                    "tx_heartbeat_errors": 0,
                    "tx_packets": 293,
                    "tx_window_errors": 0,
                },
                "wireless": {
                    "channel": 1,
                    "country": "US",
                    "frequency": 2412,
                    "mode": "access_point",
                    "noise": -95,
                    "ssid": "testnet",
                    "tx_power": 6,
                    "signal": -56,
                    "htmode": "VHT80",
                    "clients": [
                        {
                            "aid": 1,
                            "assoc": True,
                            "auth": True,
                            "authorized": True,
                            "ht": True,
                            "mac": "44:D1:FA:4B:00:01",
                            "mfp": False,
                            "preauth": False,
                            "rrm": [0, 0, 0, 0, 0],
                            "signature": "test_signature",
                            "vht": True,
                            "wds": False,
                            "wmm": True,
                            "wps": False,
                        }
                    ],
                },
                "addresses": [
                    {
                        "proto": "static",
                        "family": "ipv4",
                        "address": "192.168.1.1",
                        "mask": 8,
                    },
                    {
                        "proto": "dhcp",
                        "family": "ipv4",
                        "address": "10.0.0.4",
                        "mask": 24,
                    },
                    {
                        "proto": "static",
                        "family": "ipv6",
                        "address": "2001:3238:dfe1:ec63::fefb",
                        "mask": 12,
                    },
                    {
                        "proto": "dhcp",
                        "family": "ipv6",
                        "address": "fd46:9038:f983::1",
                        "mask": 22,
                    },
                ],
            },
        ],
        "dhcp_leases": [
            {
                "expiry": 1586943200,
                "mac": "f2:f1:3e:56:d2:77",
                "ip": "192.168.66.196",
                "client_name": "MyPhone1",
                "client_id": "01:20:f4:78:19:3b:38",
            },
            {
                "expiry": 1586943200,
                "mac": "f2:f1:3e:56:d2:77",
                "ip": "fe80::9683:c4ff:fe02:c2bf",
                "client_name": "MyPhone1",
                "client_id": "01:20:f4:78:19:3b:38",
            },
        ],
        "neighbors": [
            {
                "ip": "fe80::9683:c4ff:fe02:c2bf",
                "mac": "44:D1:FA:4B:00:02",
                "interface": "eth2",
                "state": "REACHABLE",
            },
            {
                "ip": "192.168.56.1",
                "mac": "44:D1:FA:4B:00:03",
                "interface": "br-mng",
                "state": "STALE",
            },
            {
                "ip": "192.168.56.2",
                "mac": "44:D1:FA:4B:00:04",
                "interface": "br-mng",
                "state": "STALE",
            },
        ],
    }

    def _create_device(self, **kwargs):
        if 'organization' not in kwargs:
            kwargs['organization'] = self._create_org()
        return super()._create_device(**kwargs)

    def _create_device_data(self, **kwargs):
        d = self._create_device(**kwargs)
        return DeviceData(pk=d.pk)


class TestDeviceData(BaseTestCase):
    """Test openwisp_monitoring.device.models.DeviceData"""

    def test_clean_data_ok(self):
        dd = self._create_device_data()
        dd.data = {'type': 'DeviceMonitoring', 'interfaces': []}
        dd.validate_data()

    def test_clean_sample_data_ok(self):
        dd = self._create_device_data()
        dd.data = self._sample_data
        dd.validate_data()

    def test_clean_data_fail(self):
        dd = self._create_device_data()
        try:
            dd.data = {'type': 'DeviceMonitoring', 'interfaces': [{}]}
            dd.validate_data()
        except ValidationError as e:
            self.assertIn('Invalid data in', e.message)
            self.assertIn('"#/interfaces/0"', e.message)
        else:
            self.fail('ValidationError not raised')

    def test_validate_neighbors_data(self):
        dd = self._create_device_data()
        try:
            dd.data = deepcopy(self._sample_data)
            dd.data["neighbors"][0]["ip"] = "invalid"
            dd.validate_data()
        except ValidationError as e:
            self.assertIn('Invalid data in', e.message)
            self.assertIn("is not valid under any of the given schemas", e.message)
        else:
            self.fail('ValidationError not raised')

    def test_save_data(self):
        dd = self._create_device_data()
        dd.data = deepcopy(self._sample_data)
        dd.save_data()
        return dd

    def test_read_data(self):
        dd = self.test_save_data()
        dd = DeviceData(pk=dd.pk)
        self.assertEqual(dd.data, self._sample_data)

    def test_read_data_none(self):
        dd = self._create_device_data()
        self.assertEqual(dd.data, None)

    def test_json(self):
        dd = self._create_device_data()
        dd.data = self._sample_data
        try:
            json.loads(dd.json(indent=True))
        except Exception:
            self.fail('json method did not return valid JSON')

    def test_init(self):
        dd = DeviceData()
        self.assertIsNone(dd.data)
        dd = DeviceData(data=self._sample_data)
        self.assertEqual(dd.data, self._sample_data)

    def test_device_deleted(self):
        d = self._create_device()
        metric = self._create_object_metric(name='test', content_object=d)
        metric.full_clean()
        metric.save()
        d.delete(check_deactivated=False)
        try:
            metric.refresh_from_db()
        except ObjectDoesNotExist:
            pass
        else:
            self.fail('metric was not deleted')

    def test_device_data_time_stamp(self):
        dd = self.test_save_data()
        dd = DeviceData(pk=dd.pk)
        dd.data
        self.assertIsNotNone(dd.data_timestamp)

    def test_local_time_update(self):
        dd = deepcopy(self.test_save_data())
        dd = DeviceData(pk=dd.pk)
        data = dd.data_user_friendly
        self.assertNotEqual(
            data['general']['local_time'], self._sample_data['general']['local_time']
        )

    def test_uptime_update(self):
        dd = deepcopy(self.test_save_data())
        dd = DeviceData(pk=dd.pk)
        data = dd.data_user_friendly
        self.assertNotEqual(
            data['general']['uptime'], self._sample_data['general']['uptime']
        )

    def test_wireless_interface_htmode(self):
        dd = deepcopy(self.test_save_data())
        dd = DeviceData(pk=dd.pk)
        data = dd.data_user_friendly

        with self.subTest('Test wireless interfaces with HT & VHT htmode'):
            self.assertEqual(
                data['interfaces'][0]['wireless']['htmode'], 'WiFi 4 (802.11n): HT20'
            )
            self.assertEqual(
                data['interfaces'][1]['wireless']['htmode'], 'WiFi 5 (802.11ac): VHT80'
            )

        with self.subTest('Test wireless interfaces with HE & VHT htmode'):
            test_data = deepcopy(self._sample_data)
            test_data['interfaces'][0]['wireless'].update({'htmode': 'HE40'})
            dd.data = test_data
            dd.save_data()
            data = dd.data_user_friendly
            self.assertEqual(
                data['interfaces'][0]['wireless']['htmode'], 'WiFi 6 (802.11ax): HE40'
            )
            self.assertEqual(
                data['interfaces'][1]['wireless']['htmode'], 'WiFi 5 (802.11ac): VHT80'
            )

        with self.subTest('Test wireless interfaces with NOHT & OTHER htmode'):
            test_data = deepcopy(self._sample_data)
            test_data['interfaces'][0]['wireless'].update({'htmode': 'NOHT'})
            test_data['interfaces'][1]['wireless'].update({'htmode': 'NEW_MODE'})
            dd.data = test_data
            dd.save_data()
            data = dd.data_user_friendly
            self.assertEqual(
                data['interfaces'][0]['wireless']['htmode'], 'Legacy Mode: NOHT'
            )
            self.assertEqual(
                data['interfaces'][1]['wireless']['htmode'], 'Other: NEW_MODE'
            )

    def test_bad_address_fail(self):
        dd = self._create_device_data()
        data = deepcopy(self._sample_data)
        data['interfaces'][1]['addresses'][0]['address'] = '123'
        try:
            dd.data = data
            dd.validate_data()
        except ValidationError as e:
            self.assertIn('Invalid data in', e.message)
            self.assertIn("is not valid under any of the given schemas", e.message)
        else:
            self.fail('ValidationError not raised')

    def test_bad_dhcp_lease_fail(self):
        dd = self._create_device_data()
        data = deepcopy(self._sample_data)
        data['dhcp_leases'][0]['ip'] = '123'
        try:
            dd.data = data
            dd.validate_data()
        except ValidationError as e:
            self.assertIn('Invalid data in', e.message)
            self.assertIn("is not valid under any of the given schemas", e.message)
        else:
            self.fail('ValidationError not raised')

    def test_cached_memory_optional(self):
        dd = self._create_device_data()
        data = deepcopy(self._sample_data)
        del data['resources']['memory']['cached']
        dd.data = data
        dd.validate_data()

    @patch('openwisp_monitoring.device.settings.MAC_VENDOR_DETECTION', True)
    def test_mac_vendor_info(self):
        dd = self.test_save_data()
        dd = DeviceData(pk=dd.pk)
        vendor = 'Shenzhen Yunlink Technology Co., Ltd'
        assert dd.data['interfaces']
        assert dd.data['neighbors']
        assert dd.data['dhcp_leases']
        for interface in dd.data['interfaces']:
            if 'wireless' not in interface and 'clients' not in interface['wireless']:
                continue
            for client in interface['wireless']['clients']:
                self.assertIn('vendor', client)
                self.assertEqual(client['vendor'], vendor)
        for neighbor in dd.data['neighbors']:
            self.assertIn('vendor', neighbor)
            self.assertEqual(neighbor['vendor'], vendor)
        for lease in dd.data['dhcp_leases']:
            self.assertIn('vendor', lease)

    @patch('openwisp_monitoring.device.settings.MAC_VENDOR_DETECTION', True)
    def test_mac_vendor_info_empty(self):
        dd = self._create_device_data()
        dd.data = deepcopy(self._sample_data)
        dd.data['neighbors'] = [
            {'ip': '2001:db80::1', 'interface': 'eth2.1', 'state': 'FAILED'}
        ]
        dd.save_data()
        self.assertEqual(dd.data['neighbors'][0]['vendor'], '')

    def test_bad_disk_fail(self):
        dd = self._create_device_data()
        data = deepcopy(self._sample_data)
        with self.subTest('Incorrect type'):
            data['resources']['disk'] = dict()
            try:
                dd.data = data
                dd.validate_data()
            except ValidationError as e:
                self.assertIn('Invalid data in', e.message)
                self.assertIn("{} is not of type \'array\'", e.message)
            else:
                self.fail('ValidationError not raised')
        with self.subTest('Missing required fields'):
            data['resources']['disk'] = [{'used_bytes': 18792}]
            try:
                dd.data = data
                dd.validate_data()
            except ValidationError as e:
                self.assertIn('Invalid data in', e.message)
                self.assertIn("'mount_point\' is a required property", e.message)
            else:
                self.fail('ValidationError not raised')
        with self.subTest('Incorrect field type'):
            data = deepcopy(self._sample_data)
            data['resources']['disk'][0]['used_bytes'] = 18792.12
            try:
                dd.data = data
                dd.validate_data()
            except ValidationError as e:
                self.assertIn('Invalid data in', e.message)
                self.assertIn("18792.12 is not of type \'integer\'", e.message)
            else:
                self.fail('ValidationError not raised')

    def test_resources_no_key(self):
        dd = self._create_device_data()
        with self.subTest('Test No Resources'):
            data = deepcopy(self._sample_data)
            del data['resources']
            dd.data = data
            dd.validate_data()
        with self.subTest('Test No Load'):
            data = deepcopy(self._sample_data)
            del data['resources']['load']
            dd.data = data
            dd.validate_data()
        with self.subTest('Test No Memory'):
            data = deepcopy(self._sample_data)
            del data['resources']['memory']
            dd.data = data
            dd.validate_data()
        with self.subTest('Test No Disk'):
            data = deepcopy(self._sample_data)
            del data['resources']['disk']
            dd.data = data
            dd.validate_data()

    @patch('logging.Logger.warning')
    def test_trigger_device_checks_task_resiliency(self, mock):
        dd = DeviceData(name='Test Device')
        trigger_device_checks.delay(dd.pk)
        mock.assert_called_with(f'The device with uuid {dd.pk} has been deleted')

    def test_device_data_cache_set(self):
        dd = self.create_test_data(no_resources=True)
        cache_key = get_device_cache_key(dd, context='current-data')
        cache_data = cache.get(cache_key)[0]['data']
        self.assertEqual(json.loads(cache_data), dd.data)
        with patch.object(timeseries_db, 'query', side_effect=Exception):
            dd.refresh_from_db()
            self.assertEqual(json.loads(cache_data), dd.data)

    @patch('openwisp_controller.connection.tasks.logger.info')
    def test_can_be_updated(self, mocked_logger_info):
        device = self._create_device_config()
        device.monitoring.status = 'critical'
        device.monitoring.save()
        update_config.delay(device.pk)
        mocked_logger_info.assert_called_once()

    @patch('openwisp_controller.connection.tasks.logger.info')
    def test_can_be_updated_unknown(self, mocked_logger_info):
        device = self._create_device_config()
        device.monitoring.status = 'unknown'
        device.monitoring.save()
        update_config.delay(device.pk)
        mocked_logger_info.assert_called_once()

    def test_calculate_increment(self):
        dd = self._create_device_data()
        dd.writer._init_previous_data()
        result = dd.writer._calculate_increment('wlan0', 'rx_bytes', 1234.56)
        self.assertEqual(result, 1234)


class TestDeviceMonitoring(CreateConnectionsMixin, BaseTestCase):
    """Test openwisp_monitoring.device.models.DeviceMonitoring"""

    def _create_env(self):
        d = self._create_device()
        dm = d.monitoring
        dm.status = 'ok'
        dm.save()
        ping = self._create_object_metric(
            name='ping', key='ping', field_name='reachable', content_object=d
        )
        self._create_alert_settings(
            metric=ping, custom_operator='<', custom_threshold=1, custom_tolerance=0
        )
        load = self._create_object_metric(name='load', content_object=d)
        self._create_alert_settings(
            metric=load, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        process_count = self._create_object_metric(
            name='process_count', content_object=d
        )
        self._create_alert_settings(
            metric=process_count,
            custom_operator='>',
            custom_threshold=20,
            custom_tolerance=0,
        )
        return dm, ping, load, process_count

    def test_disabling_critical_check(self):
        Check = load_model('check', 'Check')
        dm, ping, load, process_count = self._create_env()
        ping_check_instance = Check.objects.create(
            name='Ping Check',
            check_type='ping',
            content_object=dm.device,
            params={},
        )
        dm.update_status('ok')
        with catch_signal(health_status_changed) as handler:
            ping_check_instance.is_active = False
            ping_check_instance.save()
        self.assertEqual(handler.call_count, 1)
        call_args = handler.call_args[1]
        self.assertEqual(call_args['instance'], dm)
        self.assertEqual(call_args['status'], 'unknown')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'unknown')
        with self.subTest(
            'Ensure status does not change on saving active critical check'
        ):
            ping_check_instance.is_active = True
            ping_check_instance.save()
            dm.refresh_from_db()
            self.assertEqual(dm.status, 'unknown')

    def test_saving_non_critical_check(self):
        Check = load_model('check', 'Check')
        dm, ping, load, process_count = self._create_env()
        # Ensure initial status is 'ok'
        dm.update_status('ok')
        # Created a non-critical check
        non_critical_check = Check.objects.create(
            name='Configuration Applied',
            check_type='non_critical',
            content_object=dm.device,
            params={},
        )
        with catch_signal(health_status_changed) as handler:
            non_critical_check.is_active = False
            non_critical_check.save()
        self.assertEqual(
            handler.call_count, 0, 'Signal should not be fired for non-critical check'
        )
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'ok')

    def test_deleting_critical_check(self):
        Check = load_model('check', 'Check')
        dm, ping, load, process_m = self._create_env()
        ping_check_instance = Check.objects.create(
            name='Ping Check',
            check_type='ping',
            content_object=dm.device,
            params={},
        )
        dm.update_status('ok')
        with catch_signal(health_status_changed) as handler:
            ping_check_instance.delete()
        self.assertEqual(handler.call_count, 1)
        call_args = handler.call_args[1]
        self.assertEqual(call_args['instance'], dm)
        self.assertEqual(call_args['status'], 'unknown')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'unknown')

    def test_status_changed(self):
        dm, ping, load, process_count = self._create_env()
        # check signal
        with catch_signal(health_status_changed) as handler:
            dm.update_status('problem')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'problem')
        handler.assert_called_once_with(
            instance=dm,
            status='problem',
            sender=DeviceMonitoring,
            signal=health_status_changed,
        )

    def test_ok_critical_ok(self):
        dm, ping, load, process_count = self._create_env()
        self.assertEqual(dm.status, 'ok')
        ping.check_threshold(0)
        self.assertEqual(dm.status, 'critical')
        ping.check_threshold(1)
        self.assertEqual(dm.status, 'ok')

    def test_ok_problem_ok(self):
        dm, ping, load, process_count = self._create_env()
        self.assertEqual(dm.status, 'ok')
        load.check_threshold(100)
        self.assertEqual(dm.status, 'problem')
        load.check_threshold(20)
        self.assertEqual(dm.status, 'ok')

    def test_ok_problem_critical_problem_ok(self):
        dm, ping, load, process_count = self._create_env()
        self.assertEqual(dm.status, 'ok')
        load.check_threshold(100)
        self.assertEqual(dm.status, 'problem')
        ping.check_threshold(0)
        self.assertEqual(dm.status, 'critical')
        ping.check_threshold(1)
        self.assertEqual(dm.status, 'problem')
        load.check_threshold(80)
        self.assertEqual(dm.status, 'ok')

    def test_ok_critical_critical_critical_ok(self):
        dm, ping, load, process_count = self._create_env()
        self.assertEqual(dm.status, 'ok')
        ping.check_threshold(0)
        self.assertEqual(dm.status, 'critical')
        load.check_threshold(100)
        self.assertEqual(dm.status, 'critical')
        load.check_threshold(80)
        self.assertEqual(dm.status, 'critical')
        ping.check_threshold(1)
        self.assertEqual(dm.status, 'ok')

    def test_ok_problem_problem_problem_ok(self):
        dm, ping, load, process_count = self._create_env()
        self.assertEqual(dm.status, 'ok')
        load.check_threshold(100)
        self.assertEqual(dm.status, 'problem')
        process_count.check_threshold(40)
        self.assertEqual(dm.status, 'problem')
        process_count.check_threshold(10)
        self.assertEqual(dm.status, 'problem')
        load.check_threshold(80)
        self.assertEqual(dm.status, 'ok')

    def test_management_ip_clear_device_offline(self):
        dm, ping, load, process_count = self._create_env()
        dm.device.management_ip = '10.10.0.5'
        dm.device.save()
        self.assertEqual(dm.status, 'ok')
        self.assertEqual(dm.device.management_ip, '10.10.0.5')
        ping.check_threshold(0)
        self.assertEqual(dm.status, 'critical')
        self.assertIsNone(dm.device.management_ip)

    @patch('openwisp_monitoring.device.settings.AUTO_CLEAR_MANAGEMENT_IP', False)
    def test_management_ip_not_clear_device_online(self):
        dm, ping, load, process_count = self._create_env()
        dm.device.management_ip = '10.10.0.5'
        dm.device.save()
        self.assertEqual(dm.status, 'ok')
        self.assertEqual(dm.device.management_ip, '10.10.0.5')
        ping.check_threshold(0)
        self.assertEqual(dm.status, 'critical')
        self.assertIsNotNone(dm.device.management_ip)

    def _set_env_unknown(self, load, process_count, ping, dm):
        dm.status = 'unknown'
        dm.save()
        ping.is_healthy = None
        load.save()
        load.is_healthy = None
        ping.save()
        process_count.is_healthy = None
        process_count.save()

    def test_unknown_ok(self):
        dm, ping, load, process_count = self._create_env()
        self._set_env_unknown(load, process_count, ping, dm)
        load.delete()
        process_count.delete()
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'unknown')
        ping.write(1)
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'ok')

    def test_unknown_critical(self):
        dm, ping, load, process_count = self._create_env()
        self._set_env_unknown(load, process_count, ping, dm)
        load.delete()
        process_count.delete()
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'unknown')
        ping.write(0)
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'critical')

    def test_deleting_device_deletes_tsdb(self):
        dm1, ping1, _, _ = self._create_env()
        device2 = self._create_device(
            name='default.test.device2',
            mac_address='22:33:44:55:66:77',
            organization=dm1.device.organization,
        )
        dm2 = device2.monitoring
        dm2.status = 'ok'
        dm2.save()
        ping2 = self._create_object_metric(
            name='ping', key='ping', field_name='reachable', content_object=device2
        )
        ping1.write(0)
        ping2.write(0)
        self.assertNotEqual(self._read_metric(ping1), [])
        self.assertNotEqual(self._read_metric(ping2), [])
        dm1.device.delete(check_deactivated=False)
        # Only the metric related to the deleted device
        # is deleted
        self.assertEqual(self._read_metric(ping1), [])
        self.assertNotEqual(self._read_metric(ping2), [])

    def test_handle_disabled_organization(self):
        device_monitoring, _, _, _ = self._create_env()
        device = device_monitoring.device
        device.management_ip = '10.10.0.5'
        device.save()
        self.assertEqual(device_monitoring.status, 'ok')
        org = device.organization
        org.is_active = False
        org.save(update_fields=['is_active'])
        device_monitoring.refresh_from_db()
        device.refresh_from_db()
        self.assertEqual(device_monitoring.status, 'unknown')
        self.assertEqual(device.management_ip, None)

    def test_handle_deactivate_activate_device(self):
        device_monitoring, _, _, _ = self._create_env()
        device = device_monitoring.device
        self.assertEqual(device_monitoring.status, 'ok')

        with self.subTest('Test deactivation of device'):
            device.deactivate()
            device_monitoring.refresh_from_db()
            device.refresh_from_db()
            self.assertEqual(device_monitoring.status, 'deactivated')

        with self.subTest('Test activation of a deactivated device'):
            device.activate()
            device_monitoring.refresh_from_db()
            device.refresh_from_db()
            self.assertEqual(device_monitoring.status, 'unknown')


class TestWifiClientSession(TestWifiClientSessionMixin, TestCase):
    wifi_client_model = WifiClient
    wifi_session_model = WifiSession
    device_data_model = DeviceData

    def tearDown(self):
        super().tearDown()
        cache.clear()

    def test_wifi_client_session_created(self):
        data = self._sample_data
        data['interfaces'].append(self.mesh_interface)
        device_data = self._save_device_data(data=data)
        self.assertEqual(WifiClient.objects.count(), 3)
        self.assertEqual(WifiSession.objects.count(), 3)
        wifi_client1 = WifiClient.objects.get(mac_address='00:ee:ad:34:f5:3b')
        self.assertEqual(wifi_client1.vendor, None)
        self.assertEqual(wifi_client1.he, None)
        self.assertEqual(wifi_client1.vht, None)
        self.assertEqual(wifi_client1.ht, True)
        self.assertEqual(wifi_client1.wmm, True)
        self.assertEqual(wifi_client1.wds, False)
        self.assertEqual(wifi_client1.wps, False)

        wifi_client2 = WifiClient.objects.get(mac_address='b0:e1:7e:30:16:44')
        self.assertEqual(wifi_client2.vendor, None)
        self.assertEqual(wifi_client2.he, None)
        self.assertEqual(wifi_client2.vht, False)
        self.assertEqual(wifi_client2.ht, True)
        self.assertEqual(wifi_client2.wmm, True)
        self.assertEqual(wifi_client2.wds, False)
        self.assertEqual(wifi_client2.wps, False)

        wifi_client3 = WifiClient.objects.get(mac_address='c0:ee:fb:34:f5:4b')
        self.assertEqual(wifi_client3.vendor, None)
        self.assertEqual(wifi_client3.he, None)
        self.assertEqual(wifi_client3.vht, False)
        self.assertEqual(wifi_client3.ht, True)
        self.assertEqual(wifi_client3.wmm, True)
        self.assertEqual(wifi_client3.wds, False)
        self.assertEqual(wifi_client3.wps, False)

        wifi_client1_session = WifiSession.objects.get(wifi_client=wifi_client1)
        self.assertEqual(wifi_client1_session.device, device_data)
        self.assertEqual(wifi_client1_session.ssid, 'testnet')
        self.assertEqual(wifi_client1_session.interface_name, 'wlan0')
        self.assertNotEqual(wifi_client1_session.start_time, None)
        self.assertEqual(wifi_client1_session.stop_time, None)

        wifi_client2_session = WifiSession.objects.get(wifi_client=wifi_client2)
        self.assertEqual(wifi_client2_session.device, device_data)
        self.assertEqual(wifi_client2_session.ssid, 'testnet')
        self.assertEqual(wifi_client2_session.interface_name, 'wlan1')
        self.assertNotEqual(wifi_client2_session.start_time, None)
        self.assertEqual(wifi_client2_session.stop_time, None)

        wifi_client3_session = WifiSession.objects.get(wifi_client=wifi_client3)
        self.assertEqual(wifi_client3_session.device, device_data)
        self.assertEqual(wifi_client3_session.ssid, 'testnet')
        self.assertEqual(wifi_client3_session.interface_name, 'wlan1')
        self.assertNotEqual(wifi_client3_session.start_time, None)
        self.assertEqual(wifi_client3_session.stop_time, None)

    def test_updating_existing_wifi_sessions(self):
        device_data = self._create_device_data()
        self._save_device_data(device_data=device_data)
        self.assertEqual(WifiClient.objects.count(), 3)
        self.assertEqual(WifiSession.objects.count(), 3)

        self._save_device_data(device_data=device_data)
        self.assertEqual(WifiClient.objects.count(), 3)
        self.assertEqual(WifiSession.objects.count(), 3)

    def test_opening_closing_existing_sessions(self):
        data = deepcopy(self._sample_data)
        device_data = self._create_device_data()
        self._save_device_data(device_data, data)
        self.assertEqual(WifiSession.objects.filter(stop_time=None).count(), 3)
        self.assertEqual(WifiSession.objects.count(), 3)
        self.assertEqual(WifiClient.objects.count(), 3)

        with self.subTest('Test closing session'):
            self._save_device_data(
                device_data, data={'type': 'DeviceMonitoring', 'interfaces': []}
            )
            self.assertEqual(WifiSession.objects.filter(stop_time=None).count(), 0)
            self.assertEqual(WifiSession.objects.count(), 3)
            self.assertEqual(WifiClient.objects.count(), 3)

        with self.subTest('Test re-opening session for exising clients'):
            data = deepcopy(self._sample_data)
            self._save_device_data(device_data, data)
            self.assertEqual(WifiSession.objects.filter(stop_time=None).count(), 3)
            self.assertEqual(WifiSession.objects.count(), 6)
            self.assertEqual(WifiClient.objects.count(), 3)

    def test_delete_old_wifi_sessions(self):
        with freeze_time(now() - timedelta(days=6 * 31)):
            self._create_wifi_session()
        self.assertEqual(WifiSession.objects.count(), 1)
        delete_wifi_clients_and_sessions.delay()
        self.assertEqual(WifiSession.objects.count(), 0)

    def test_database_queries(self):
        device_data = self._create_device_data()

        with self.subTest('Test creating new clients and sessions'):
            data = deepcopy(self._sample_data)
            with self.assertNumQueries(28):
                self._save_device_data(device_data, data)

        with self.subTest('Test updating existing clients and sessions'):
            data = deepcopy(self._sample_data)
            with self.assertNumQueries(7):
                self._save_device_data(device_data, data)

        with self.subTest('Test closing existing sessions'):
            data = deepcopy(self._sample_data)
            with self.assertNumQueries(1):
                self._save_device_data(
                    device_data, data={'type': 'DeviceMonitoring', 'interface': []}
                )

        with self.subTest('Test new sessions for existing clients'):
            data = deepcopy(self._sample_data)
            with self.assertNumQueries(13):
                self._save_device_data(device_data, data)

    @patch.object(app_settings, 'WIFI_SESSIONS_ENABLED', False)
    def test_disabling_wifi_sessions(self):
        device_data = self._create_device_data()
        with self.assertNumQueries(0):
            self._save_device_data(device_data)
        self.assertEqual(WifiClient.objects.count(), 0)
        self.assertEqual(WifiSession.objects.count(), 0)

    def test_device_offline_close_session(self):
        start_time = now()
        device_monitoring = self._create_device_monitoring()
        ping = self._create_object_metric(
            name='ping',
            key='ping',
            field_name='reachable',
            content_object=device_monitoring.device,
        )
        ping_alerts = self._create_alert_settings(
            metric=ping, custom_operator='<', custom_threshold=1, custom_tolerance=1
        )
        load = self._create_object_metric(
            name='load', content_object=device_monitoring.device
        )
        self._create_alert_settings(
            metric=load, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        ping.write(1)
        load.write(50)
        wifi_client = self._create_wifi_client()
        self._create_wifi_session(
            wifi_client=wifi_client, device=device_monitoring.device
        )

        def _assert_open_wifi_session(count):
            self.assertEqual(
                WifiSession.objects.filter(stop_time__isnull=True).count(), count
            )
            self.assertEqual(WifiSession.objects.count(), 1)

        # Sanity check
        _assert_open_wifi_session(1)

        # WiFi session is not closed when a non-critical metric trepasses the threshold
        load.write(95)
        _assert_open_wifi_session(1)

        # WiFi session is not closed when a critical metric trepasses the threshold
        # within the tolerance limit
        ping.write(0)
        _assert_open_wifi_session(1)

        # WiFi session is closed when a critical metric trepasses the threshold
        # beyond the tolerance limit
        with freeze_time(start_time + timedelta(minutes=2)):
            ping.write(0)
        _assert_open_wifi_session(0)

        WifiSession.objects.update(stop_time=None)
        # Reset ping.is_healthy and ping.is_healthy_tolerant
        # Sets ping.is_healthy to True
        with freeze_time(start_time + timedelta(minutes=2)):
            ping.write(1)
        # Sets ping.is_healthy_tolerant to True
        with freeze_time(start_time + timedelta(minutes=3)):
            ping.write(1)
        _assert_open_wifi_session(1)

        # WiFi session is closed when when a critical metric trepasses the threshold
        # beyond the tolerance limit when alerts are turned off
        ping_alerts.is_active = False
        ping_alerts.save()
        # Sets ping.is_healthy to False, doesn't closes WiFi Session
        with freeze_time(start_time + timedelta(minutes=4)):
            ping.write(0)
        _assert_open_wifi_session(1)
        # Sets ping.is_healthy_tolerant to False, closes WiFi Session
        with freeze_time(start_time + timedelta(minutes=5)):
            ping.write(0)
        _assert_open_wifi_session(0)



from django.conf import settings

from ..settings import get_settings_value

CHECK_CLASSES = get_settings_value(
    'CHECK_CLASSES',
    (
        ('openwisp_monitoring.check.classes.Ping', 'Ping'),
        ('openwisp_monitoring.check.classes.ConfigApplied', 'Configuration Applied'),
        ('openwisp_monitoring.check.classes.Iperf3', 'Iperf3'),
    ),
)
AUTO_PING = get_settings_value('AUTO_PING', True)
AUTO_CONFIG_CHECK = get_settings_value('AUTO_DEVICE_CONFIG_CHECK', True)
# If OPENWISP_MONITORING_MANAGEMENT_IP_ONLY is not configured, use
# OPENWISP_CONTROLLER_MANAGEMENT_IP_ONLY.
MANAGEMENT_IP_ONLY = get_settings_value(
    'MANAGEMENT_IP_ONLY',
    getattr(settings, 'OPENWISP_CONTROLLER_MANAGEMENT_IP_ONLY', True),
)
PING_CHECK_CONFIG = get_settings_value('PING_CHECK_CONFIG', {})
AUTO_IPERF3 = get_settings_value('AUTO_IPERF3', False)
IPERF3_CHECK_CONFIG = get_settings_value('IPERF3_CHECK_CONFIG', {})
IPERF3_CHECK_LOCK_EXPIRE = get_settings_value(
    'IPERF3_CHECK_LOCK_EXPIRE', 10 * 60
)  # 10 minutes arbitrarily chosen (must be longer than TCP + UDP test time)
IPERF3_CHECK_DELETE_RSA_KEY = get_settings_value('IPERF3_CHECK_DELETE_RSA_KEY', True)
CHECKS_LIST = get_settings_value('CHECK_LIST', list(dict(CHECK_CLASSES).keys()))
CONFIG_CHECK_INTERVAL = int(
    get_settings_value('CONFIG_CHECK_INTERVAL', 5)
)  # in minutes



from django.contrib.auth.models import Permission

from openwisp_controller.migrations import create_default_permissions, get_swapped_model


def assign_check_inline_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)
    operators_read_only_admins_manage = [
        'check',
    ]
    manage_operations = ['add', 'view', 'change', 'delete']
    Group = get_swapped_model(apps, 'openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    for model_name in operators_read_only_admins_manage:
        try:
            permission = Permission.objects.get(
                codename='view_{}_inline'.format(model_name)
            )
            operator.permissions.add(permission.pk)
        except Permission.DoesNotExist:
            pass
        for operation in manage_operations:
            permission = Permission.objects.get(
                codename='{}_{}_inline'.format(operation, model_name)
            )
            admin.permissions.add(permission.pk)



from unittest.mock import patch

from django.core import management
from django.test import TransactionTestCase
from swapper import load_model

from ...device.tests import TestDeviceMonitoringMixin
from ..classes import Ping
from ..settings import CHECK_CLASSES
from ..tasks import perform_check
from ..utils import run_checks_async
from . import _FPING_REACHABLE

Check = load_model('check', 'Check')


class TestUtils(TestDeviceMonitoringMixin, TransactionTestCase):
    _PING = CHECK_CLASSES[0][0]

    def _create_check(self):
        device = self._create_device(organization=self._create_org())
        device.last_ip = '10.40.0.1'
        device.save()
        # check is automatically created via django signal

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def test_run_checks_async_success(self, mocked_method):
        self._create_check()
        run_checks_async()

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def test_management_command(self, mocked_method):
        self._create_check()
        management.call_command('run_checks')

    @patch('logging.Logger.warning')
    def test_perform_check_task_resiliency(self, mock):
        check = Check(name='Test check')
        perform_check.delay(check.pk)
        mock.assert_called_with(f'The check with uuid {check.pk} has been deleted')



# Generated by Django 4.0.4 on 2022-05-17 11:28

from django.db import migrations

from . import assign_permissions_to_groups


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0004_wificlient_wifisession'),
    ]

    operations = [
        migrations.RunPython(
            assign_permissions_to_groups, reverse_code=migrations.RunPython.noop
        ),
    ]



from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase


class TestManagementCommands(TestCase):
    @patch('openwisp_monitoring.monitoring.tasks.migrate_timeseries_database.delay')
    def test_migrate_timeseries(self, mocked_celery_task):
        call_command('migrate_timeseries')
        mocked_celery_task.assert_called_once()



# Generated by Django 3.2.15 on 2022-09-08 11:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('monitoring', '0010_add_alertsettings_inline_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metric',
            name='field_name',
            field=models.CharField(
                blank=True,
                default='value',
                help_text='leave blank to determine automatically',
                max_length=16,
            ),
        ),
    ]



import json
from unittest.mock import patch

from django.core.cache import cache
from django.urls import reverse
from swapper import load_model

from ..signals import health_status_changed
from ..tasks import trigger_device_checks
from ..utils import get_device_cache_key
from . import DeviceMonitoringTestCase, DeviceMonitoringTransactionTestcase

DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
Device = load_model('config', 'Device')


class TestRecovery(DeviceMonitoringTestCase):
    """Tests ``Device Recovery Detection`` functionality"""

    def test_device_recovery_cache_key_not_set(self):
        device_monitoring_app = DeviceMonitoring._meta.app_config
        health_status_changed.disconnect(
            device_monitoring_app.manage_device_recovery_cache_key,
            sender=DeviceMonitoring,
            dispatch_uid='recovery_health_status_changed',
        )
        with patch(
            'openwisp_monitoring.device.apps.app_settings.DEVICE_RECOVERY_DETECTION',
            False,
        ):
            device_monitoring_app.device_recovery_detection()
        dm = self._create_device_monitoring()
        cache_key = get_device_cache_key(device=dm.device)
        dm.update_status('critical')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'critical')
        self.assertEqual(cache.get(cache_key), None)
        device_monitoring_app.device_recovery_detection()

    def test_device_recovery_cache_key_set(self):
        dm = self._create_device_monitoring()
        cache_key = get_device_cache_key(device=dm.device)
        dm.update_status('critical')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'critical')
        self.assertEqual(cache.get(cache_key), 1)
        dm.update_status('problem')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'problem')
        self.assertEqual(cache.get(cache_key), None)
        dm.update_status('ok')
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'ok')
        self.assertEqual(cache.get(cache_key), None)

    def test_status_set_ok(self):
        """Tests device status is set to ok if no related checks present"""
        dm = self._create_device_monitoring()
        dm.update_status('critical')
        trigger_device_checks.delay(dm.device.pk)
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'ok')

    def test_status_set_critical(self):
        """Tests device status is set to critical if no related checks present and recovery=False is passed"""
        dm = self._create_device_monitoring()
        dm.update_status('critical')
        trigger_device_checks.delay(dm.device.pk, recovery=False)
        dm.refresh_from_db()
        self.assertEqual(dm.status, 'critical')


class TestRecoveryTransaction(DeviceMonitoringTransactionTestcase):
    @patch('openwisp_monitoring.device.tasks.trigger_device_checks.delay')
    def test_device_checksum_trigger_device_recovery_checks(self, mocked_task):
        device = self._create_config(organization=self._get_org()).device
        device_monitoring = device.monitoring

        with self.subTest('Test checks are not triggered if device is "ok"'):
            device_monitoring.update_status('ok')
            response = self.client.get(
                reverse('controller:device_checksum', args=[device.pk]),
                data={'key': device.key},
            )
            self.assertEqual(response.status_code, 200)
            mocked_task.assert_not_called()

        mocked_task.reset_mock()

        with self.subTest('Test checks are triggered if device is "critical"'):
            device_monitoring.update_status('critical')
            response = self.client.get(
                reverse('controller:device_checksum', args=[device.pk]),
                data={'key': device.key},
            )
            self.assertEqual(response.status_code, 200)
            mocked_task.assert_called_once()

    @patch('openwisp_monitoring.device.tasks.trigger_device_checks.delay')
    def test_metrics_received_trigger_device_recovery_checks(self, mocked_task):
        device = self._create_config(organization=self._get_org()).device
        device_monitoring = device.monitoring
        netjson = json.dumps(self._data())
        url = self._url(device.pk, key=device.key)

        with self.subTest('Test checks are not triggered if device is "ok"'):
            device_monitoring.update_status('ok')

            response = self.client.post(
                url,
                data=netjson,
                content_type='application/json',
            )
            self.assertEqual(response.status_code, 200)
            mocked_task.assert_not_called()

        mocked_task.reset_mock()

        with self.subTest('Test checks are triggered if device is "critical"'):
            device_monitoring.update_status('critical')
            response = self.client.post(
                url,
                data=netjson,
                content_type='application/json',
            )
            self.assertEqual(response.status_code, 200)
            mocked_task.assert_called_once()



import csv
import logging
from collections import OrderedDict
from datetime import datetime
from io import StringIO

from django.conf import settings
from django.http import HttpResponse
from pytz import timezone
from pytz import timezone as tz
from pytz.exceptions import UnknownTimeZoneError
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from swapper import load_model

from .monitoring.exceptions import InvalidChartConfigException

logger = logging.getLogger(__name__)

Chart = load_model('monitoring', 'Chart')


class MonitoringApiViewMixin:
    def _get_charts(self, request, *args, **kwargs):
        """Hook to return Chart query."""
        raise NotImplementedError

    def _get_additional_data(request, *args, **kwargs):
        """Hook to return any additonal data that should be included in the response."""
        return {}

    def _validate_custom_date(self, start, end, tmz):
        try:
            start = datetime.strptime(start, '%Y-%m-%d %H:%M:%S')
            end = datetime.strptime(end, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            raise ValidationError(
                'Incorrect custom date format, should be YYYY-MM-DD H:M:S'
            )
        if (end - start).days > 365:
            raise ValidationError("The date range shouldn't be greater than 365 days")
        if start > end:
            raise ValidationError('start_date cannot be greater than end_date')
        now_tz = datetime.now(tz=timezone(tmz)).strftime('%Y-%m-%d %H:%M:%S')
        now = datetime.strptime(now_tz, '%Y-%m-%d %H:%M:%S')
        if start > now:
            raise ValidationError("start_date cannot be greater than today's date")
        if end > now:
            raise ValidationError("end_date cannot be greater than today's date")
        return start, end

    def get(self, request, *args, **kwargs):
        time = request.query_params.get('time', Chart.DEFAULT_TIME)
        start_date = request.query_params.get('start', None)
        end_date = request.query_params.get('end', None)
        # try to read timezone
        timezone = request.query_params.get('timezone', settings.TIME_ZONE)
        try:
            tz(timezone)
        except UnknownTimeZoneError:
            raise ValidationError('Unkown Time Zone')
        # if custom dates are provided then validate custom dates
        if start_date and end_date:
            start_datetime, end_datetime = self._validate_custom_date(
                start_date, end_date, timezone
            )
            # if valid custom dates then calculate custom days
            time = '1d'
            custom_days = (end_datetime - start_datetime).days
            if custom_days:
                time = f'{custom_days}d'
        if time not in Chart._get_group_map(time).keys():
            raise ValidationError('Time range not supported')
        charts = self._get_charts(request, *args, **kwargs)
        # prepare response data
        data = self._get_charts_data(charts, time, timezone, start_date, end_date)
        # csv export has a different response
        if request.query_params.get('csv'):
            response = HttpResponse(self._get_csv(data), content_type='text/csv')
            response['Content-Disposition'] = 'attachment; filename=data.csv'
            return response
        data.update(self._get_additional_data(request, *args, **kwargs))
        return Response(data)

    def _get_chart_additional_query_kwargs(self, chart):
        """Hook to provide additional kwargs to Chart.read."""
        return None

    def _get_charts_data(self, charts, time, timezone, start_date, end_date):
        chart_map = {}
        x_axys = True
        data = OrderedDict({'charts': []})
        for chart in charts:
            # prepare chart dict
            try:
                chart_dict = chart.read(
                    time=time,
                    x_axys=x_axys,
                    timezone=timezone,
                    start_date=start_date,
                    end_date=end_date,
                    additional_query_kwargs=self._get_chart_additional_query_kwargs(
                        chart
                    ),
                )
                if not chart_dict['traces']:
                    continue
                chart_dict['description'] = chart.description
                chart_dict['title'] = chart.title.format(
                    metric=chart.metric, **chart.metric.tags
                )
                chart_dict['type'] = chart.type
                chart_dict['unit'] = chart.unit
                chart_dict['summary_labels'] = chart.summary_labels
                chart_dict['colors'] = chart.colors
                chart_dict['colorscale'] = chart.colorscale
                for attr in ['fill', 'xaxis', 'yaxis']:
                    value = getattr(chart, attr)
                    if value:
                        chart_dict[attr] = value
                if chart.trace_type:
                    chart_dict['trace_type'] = chart.trace_type
                if chart.trace_order:
                    chart_dict['trace_order'] = chart.trace_order
                if chart.calculate_total:
                    chart_dict['calculate_total'] = chart.calculate_total
                if chart.connect_points:
                    chart_dict['connect_points'] = chart.connect_points
                if chart.trace_labels:
                    chart_dict['trace_labels'] = chart.trace_labels
            except InvalidChartConfigException:
                logger.exception(f'Skipped chart for metric {chart.metric}')
                continue
            # get x axys (only once)
            if x_axys and chart_dict['x'] and chart.type != 'histogram':
                data['x'] = chart_dict.pop('x')
                x_axys = False
            # prepare to sort the items according to
            # the order in the chart configuration
            key = f'{chart.order} {chart_dict["title"]}'
            chart_map[key] = chart_dict
        # add sorted chart list to chart data
        data['charts'] = list(OrderedDict(sorted(chart_map.items())).values())
        return data

    def _get_csv(self, data):
        header = ['time']
        columns = [data.get('x')]
        histograms = []
        for chart in data['charts']:
            if chart['type'] == 'histogram':
                histograms.append(chart)
                continue
            for trace in chart['traces']:
                header.append(self._get_csv_header(chart, trace))
                columns.append(trace[1])
        rows = [header]
        for index, element in enumerate(data.get('x', [])):
            row = []
            for column in columns:
                row.append(column[index])
            rows.append(row)
        for chart in histograms:
            rows.append([])
            rows.append([chart['title']])
            # Export value as 0 if it is None
            for key, value in chart['summary'].items():
                if chart['summary'][key] is None:
                    chart['summary'][key] = 0
            # Sort Histogram on the basis of value in the descending order
            sorted_charts = sorted(
                chart['summary'].items(), key=lambda x: x[1], reverse=True
            )
            for field, value in sorted_charts:
                rows.append([field, value])
        # write CSV to in-memory file object
        fileobj = StringIO()
        csv.writer(fileobj).writerows(rows)
        return fileobj.getvalue()

    def _get_csv_header(self, chart, trace):
        header = trace[0]
        return f'{header} - {chart["title"]}'



# Generated by Django 4.0.4 on 2022-06-03 11:36

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0007_add_wificlient_field_he'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='wificlient',
            options={'ordering': ('-created',), 'verbose_name': 'WiFi Client'},
        ),
    ]



import logging
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DatabaseError

logger = logging.getLogger(__name__)

TIMESERIES_DB = getattr(settings, 'TIMESERIES_DATABASE', None)
if not TIMESERIES_DB:
    TIMESERIES_DB = {
        'BACKEND': 'openwisp_monitoring.db.backends.influxdb',
        'USER': getattr(settings, 'INFLUXDB_USER', 'openwisp'),
        'PASSWORD': getattr(settings, 'INFLUXDB_PASSWORD', 'openwisp'),
        'NAME': getattr(settings, 'INFLUXDB_DATABASE', 'openwisp2'),
        'HOST': getattr(settings, 'INFLUXDB_HOST', 'localhost'),
        'PORT': getattr(settings, 'INFLUXDB_PORT', '8086'),
    }
    logger.warning(
        'The previous method to define Timeseries Database has been deprecated. Please refer to the docs:\n'
        'https://github.com/openwisp/openwisp-monitoring#setup-integrate-in-an-existing-django-project'
    )


def load_backend_module(backend_name=TIMESERIES_DB['BACKEND'], module=None):
    """Loads backend module.

    Returns database backend module given a fully qualified database
    backend name, or raise an error if it doesn't exist or backend is not
    well defined.
    """
    try:
        assert 'BACKEND' in TIMESERIES_DB, 'BACKEND'
        assert 'USER' in TIMESERIES_DB, 'USER'
        assert 'PASSWORD' in TIMESERIES_DB, 'PASSWORD'
        assert 'NAME' in TIMESERIES_DB, 'NAME'
        assert 'HOST' in TIMESERIES_DB, 'HOST'
        assert 'PORT' in TIMESERIES_DB, 'PORT'
        if module:
            return import_module(f'{backend_name}.{module}')
        else:
            return import_module(backend_name)
    except AttributeError as e:
        raise DatabaseError('No TIMESERIES_DATABASE specified in settings') from e
    except AssertionError as e:
        raise ImproperlyConfigured(
            f'"{e}" field is not declared in TIMESERIES_DATABASE'
        ) from e
    except ImportError as e:
        # The database backend wasn't found. Display a helpful error message
        # listing all built-in database backends.
        builtin_backends = ['influxdb']
        if backend_name not in [
            f'openwisp_monitoring.db.backends.{b}' for b in builtin_backends
        ]:
            raise ImproperlyConfigured(
                f"{backend_name} isn't an available database backend.\n"
                "Try using 'openwisp_monitoring.db.backends.XXX', where XXX is one of:\n"
                f"{builtin_backends}"
            ) from e


timeseries_db = load_backend_module(module='client').DatabaseClient()
timeseries_db.queries = load_backend_module(module='queries')



from datetime import timedelta

from django.utils import timezone
from swapper import load_model

from ...device.utils import SHORT_RP
from ..settings import CONFIG_CHECK_INTERVAL
from .base import BaseCheck

AlertSettings = load_model('monitoring', 'AlertSettings')


class ConfigApplied(BaseCheck):
    def check(self, store=True):
        # If the device is down or does not have a config
        # do not run config applied check
        if self.related_object.monitoring.status in [
            'critical',
            'unknown',
        ] or not hasattr(self.related_object, 'config'):
            return
        result = int(
            self.related_object.config.status == 'applied'
            or self.related_object.modified
            > timezone.now() - timedelta(minutes=CONFIG_CHECK_INTERVAL)
        )
        # If the device config is in error status we don't need to notify
        # the user (because that's already done by openwisp-controller)
        # but we need to ensure health status will be changed
        send_alert = self.related_object.config.status != 'error'
        if store:
            self._get_metric().write(
                result, retention_policy=SHORT_RP, send_alert=send_alert
            )
        return result

    def _get_metric(self):
        metric, created = self._get_or_create_metric(configuration='config_applied')
        if created:
            self._create_alert_setting(metric)
        return metric

    def _create_alert_setting(self, metric):
        alert_s = AlertSettings(metric=metric)
        alert_s.full_clean()
        alert_s.save()



import swapper
from django.contrib.auth.models import Permission

from openwisp_controller.migrations import create_default_permissions, get_swapped_model


def assign_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)
    operators_read_only_admins_manage = [
        'check',
        'alertsettings',
        'wificlient',
        'wifisession',
    ]
    manage_operations = ['add', 'change', 'delete']
    Group = get_swapped_model(apps, 'openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    for model_name in operators_read_only_admins_manage:
        try:
            permission = Permission.objects.get(codename='view_{}'.format(model_name))
            operator.permissions.add(permission.pk)
        except Permission.DoesNotExist:
            pass
        for operation in manage_operations:
            admin.permissions.add(
                Permission.objects.get(
                    codename='{}_{}'.format(operation, model_name)
                ).pk
            )


def assign_alertsettings_inline_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)
    operators_read_only_admins_manage = [
        'alertsettings',
    ]
    manage_operations = ['add', 'view', 'change', 'delete']
    Group = get_swapped_model(apps, 'openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    for model_name in operators_read_only_admins_manage:
        try:
            permission = Permission.objects.get(
                codename='view_{}_inline'.format(model_name)
            )
            operator.permissions.add(permission.pk)
        except Permission.DoesNotExist:
            pass
        for operation in manage_operations:
            permission = Permission.objects.get(
                codename='{}_{}_inline'.format(operation, model_name)
            )
            admin.permissions.add(permission.pk)


def create_general_metrics(apps, schema_editor):
    Chart = swapper.load_model('monitoring', 'Chart')
    Metric = swapper.load_model('monitoring', 'Metric')

    metric, created = Metric._get_or_create(
        configuration='general_clients',
        name='General Clients',
        key='wifi_clients',
        object_id=None,
        content_type_id=None,
    )
    if created:
        chart = Chart(metric=metric, configuration='gen_wifi_clients')
        chart.full_clean()
        chart.save()

    metric, created = Metric._get_or_create(
        configuration='general_traffic',
        name='General Traffic',
        key='traffic',
        object_id=None,
        content_type_id=None,
    )
    if created:
        chart = Chart(metric=metric, configuration='general_traffic')
        chart.full_clean()
        chart.save()


def delete_general_metrics(apps, schema_editor):
    Metric = apps.get_model('monitoring', 'Metric')
    Metric.objects.filter(content_type__isnull=True, object_id__isnull=True).delete()



from django.db import migrations
from swapper import load_model


def update_critical_device_metric_status(apps, schema_editor):
    Check = apps.get_model('check', 'Check')
    # We need to load the real concrete model here, because we
    # will be calling one of its class methods below
    DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
    critical_metrics_keys = DeviceMonitoring._get_critical_metric_keys()

    for check in Check.objects.filter(
        is_active=False, check_type__in=critical_metrics_keys
    ).iterator():
        DeviceMonitoring.handle_critical_metric(check)


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0008_alter_wificlient_options'),
    ]

    operations = [
        migrations.RunPython(
            update_critical_device_metric_status, reverse_code=migrations.RunPython.noop
        ),
    ]



import logging
from functools import wraps
from time import sleep

from django.db import transaction

from .settings import MONITORING_TIMESERIES_RETRY_OPTIONS

logger = logging.getLogger(__name__)


def transaction_on_commit(func):
    with transaction.atomic():
        transaction.on_commit(func)


def retry(method):
    @wraps(method)
    def wrapper(*args, **kwargs):
        max_retries = MONITORING_TIMESERIES_RETRY_OPTIONS.get('max_retries')
        delay = MONITORING_TIMESERIES_RETRY_OPTIONS.get('delay')
        for attempt_no in range(1, max_retries + 1):
            try:
                return method(*args, **kwargs)
            except Exception as err:
                logger.info(
                    f'Error while executing method "{method.__name__}":\n{err}\n'
                    f'Attempt {attempt_no} out of {max_retries}.\n'
                )
                if attempt_no > 3:
                    sleep(delay)
                if attempt_no == max_retries:
                    raise err

    return wrapper



# Generated by Django 3.1.2 on 2020-11-18 12:57

import uuid

import django.db.migrations.operations.special
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import swapper
from django.db import migrations, models
from django.utils.translation import gettext_lazy as _

from .. import settings as app_settings
from ..models import DeviceMonitoring


def create_device_monitoring(apps, schema_editor):
    """Data migration"""
    Device = apps.get_model('config', 'Device')
    DeviceMonitoring = apps.get_model('device_monitoring', 'DeviceMonitoring')
    for device in Device.objects.iterator():
        DeviceMonitoring.objects.create(device=device)


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        swapper.dependency('config', 'Device', '0004_add_device_model'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeviceData',
            fields=[],
            options={
                'indexes': [],
                'proxy': True,
                'swappable': swapper.swappable_setting(
                    'device_monitoring', 'DeviceData'
                ),
            },
            bases=(swapper.get_model_name('config', 'Device'),),
        ),
        migrations.CreateModel(
            name='DeviceMonitoring',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'status',
                    model_utils.fields.StatusField(
                        choices=[
                            (
                                'unknown',
                                _(app_settings.HEALTH_STATUS_LABELS['unknown']),
                            ),
                            ('ok', _(app_settings.HEALTH_STATUS_LABELS['ok'])),
                            (
                                'problem',
                                _(app_settings.HEALTH_STATUS_LABELS['problem']),
                            ),
                            (
                                'critical',
                                _(app_settings.HEALTH_STATUS_LABELS['critical']),
                            ),
                            (
                                'deactivated',
                                _(app_settings.HEALTH_STATUS_LABELS['deactivated']),
                            ),
                        ],
                        default='unknown',
                        db_index=True,
                        help_text=DeviceMonitoring._meta.get_field('status').help_text,
                        max_length=100,
                        no_check_for_status=True,
                        verbose_name='health status',
                    ),
                ),
                (
                    'device',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='monitoring',
                        to=swapper.get_model_name('config', 'Device'),
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': swapper.swappable_setting(
                    'device_monitoring', 'DeviceMonitoring'
                ),
            },
        ),
        migrations.RunPython(
            code=create_device_monitoring, reverse_code=migrations.RunPython.noop
        ),
    ]



from django.contrib import admin
from swapper import load_model

from openwisp_utils.admin import TimeReadonlyAdminMixin

Check = load_model('check', 'Check')


@admin.register(Check)
class CheckAdmin(TimeReadonlyAdminMixin, admin.ModelAdmin):
    list_display = ['__str__', 'check_type', 'created', 'modified']
    search_fields = ['name', 'object_id']
    # TODO: filters
    save_on_top = True



from collections import OrderedDict
from copy import deepcopy

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _
from openwisp_notifications.types import (
    register_notification_type,
    unregister_notification_type,
)

from openwisp_monitoring.db import chart_query
from openwisp_monitoring.monitoring.utils import clean_timeseries_data_key
from openwisp_utils.utils import deep_merge_dicts

from . import settings as app_settings

DEFAULT_COLORS = [
    '#1f77b4',  # muted blue
    '#ff7f0e',  # safety orange
    '#2ca02c',  # cooked asparagus green
    '#d62728',  # brick red
    '#9467bd',  # muted purple
    '#8c564b',  # chestnut brown
    '#e377c2',  # raspberry yogurt pink
    '#7f7f7f',  # middle gray
    '#bcbd22',  # curry yellow-green
    '#17becf',  # blue-teal
]

ACCESS_TECHNOLOGIES = OrderedDict(
    (
        ('gsm', '#b42a0c'),
        ('cdma1x', '#dd5817'),
        ('evdo', '#df7514'),
        ('umts', '#efdd50'),
        ('lte', '#67c368'),
        ('5g', '#377873'),
    )
)


def _get_access_tech():
    res = []
    keys_list = list(ACCESS_TECHNOLOGIES.keys())
    for i in reversed(keys_list):
        res.append([keys_list.index(i), ACCESS_TECHNOLOGIES[i], i])
    return res


# under discussion
DEFAULT_METRICS = {
    'ping': {
        'label': _('Ping'),
        'name': 'Ping',
        'key': 'ping',
        'field_name': 'reachable',
        'related_fields': ['loss', 'rtt_min', 'rtt_max', 'rtt_avg'],
        'charts': {
            'uptime': {
                'type': 'bar',
                'title': _('Ping Success Rate'),
                'description': _(
                    'A value of 100% means reachable, 0% means unreachable, values in '
                    'between 0% and 100% indicate the average reachability in the '
                    'period observed. Obtained with the fping linux program.'
                ),
                'summary_labels': [_('Average Ping Success Rate')],
                'unit': '%',
                'order': 200,
                'colorscale': {
                    'max': 100,
                    'min': 0,
                    'label': _('Rate'),
                    'scale': [
                        [0, '#c13000'],
                        [0.1, '#ec7204'],
                        [0.5, '#deed0e'],
                        [0.9, '#90d000'],
                        [1, '#7db201'],
                    ],
                    'map': [
                        [100, '#7db201', _('Flawless')],
                        [90, '#90d000', _('Mostly Reachable')],
                        [50, '#deed0e', _('Partly Reachable')],
                        [10, '#ec7204', _('Mostly Unreachable')],
                        [None, '#c13000', _('Unreachable')],
                    ],
                    'fixed_value': 100,
                },
                'query': chart_query['uptime'],
            },
            'packet_loss': {
                'type': 'bar',
                'title': _('Packet loss'),
                'description': _(
                    'Indicates the percentage of lost packets observed in ICMP probes. '
                    'Obtained with the fping linux program.'
                ),
                'summary_labels': [_('Average packet loss')],
                'unit': '%',
                'colors': [DEFAULT_COLORS[3]],
                'order': 210,
                'query': chart_query['packet_loss'],
            },
            'rtt': {
                'type': 'scatter',
                'title': _('Round Trip Time'),
                'description': _(
                    'Round trip time observed in ICMP probes, measuered in milliseconds.'
                ),
                'summary_labels': [
                    _('Average RTT'),
                    _('Average Max RTT'),
                    _('Average Min RTT'),
                ],
                'unit': _(' ms'),
                'order': 220,
                'query': chart_query['rtt'],
            },
        },
        'alert_settings': {'operator': '<', 'threshold': 1, 'tolerance': 0},
        'notification': {
            'problem': {
                'verbose_name': 'Ping PROBLEM',
                'verb': _('is not reachable'),
                'level': 'warning',
                'email_subject': _(
                    '[{site.name}] PROBLEM: {notification.target} {notification.verb}'
                ),
                'message': _(
                    'The device [{notification.target}]({notification.target_link}) '
                    '{notification.verb}.'
                ),
            },
            'recovery': {
                'verbose_name': 'Ping RECOVERY',
                'verb': _('is reachable again'),
                'level': 'info',
                'email_subject': _(
                    '[{site.name}] RECOVERY: {notification.target} {notification.verb}'
                ),
                'message': _(
                    'The device [{notification.target}]({notification.target_link}) '
                    '{notification.verb}.'
                ),
            },
        },
    },
    'config_applied': {
        'label': _('Configuration Applied'),
        'name': 'Configuration Applied',
        'key': 'config_applied',
        'field_name': 'config_applied',
        'alert_settings': {'operator': '<', 'threshold': 1, 'tolerance': 5},
        'notification': {
            'problem': {
                'verbose_name': 'Configuration Applied PROBLEM',
                'verb': _('has not been applied'),
                'level': 'warning',
                'email_subject': _(
                    '[{site.name}] PROBLEM: {notification.target} configuration '
                    'status issue'
                ),
                'message': _(
                    'The configuration of device [{notification.target}]'
                    '({notification.target_link}) {notification.verb} in a timely manner.'
                ),
            },
            'recovery': {
                'verbose_name': 'Configuration Applied RECOVERY',
                'verb': _('configuration has been applied again'),
                'level': 'info',
                'email_subject': _(
                    '[{site.name}] RECOVERY: {notification.target} {notification.verb} '
                    'successfully'
                ),
                'message': _(
                    'The configuration of device [{notification.target}]({notification.target_link}) '
                    '{notification.verb} successfully.'
                ),
            },
        },
    },
    'traffic': {
        'label': _('Traffic'),
        'name': '{name}',
        'key': '{key}',
        'field_name': 'rx_bytes',
        'related_fields': ['tx_bytes'],
        'charts': {
            'traffic': {
                'type': 'stackedbar+lines',
                'calculate_total': True,
                'trace_type': {
                    'download': 'stackedbar',
                    'upload': 'stackedbar',
                    'total': 'lines',
                },
                'trace_order': ['total', 'download', 'upload'],
                'title': _('Traffic: {ifname}'),
                'label': _('Traffic'),
                'description': _(
                    'Network traffic (total, download and upload) '
                    'of the interface "{ifname}".'
                ),
                'summary_labels': [
                    _('Total traffic'),
                    _('Total download traffic'),
                    _('Total upload traffic'),
                ],
                'unit': 'adaptive_prefix+B',
                'order': 240,
                'query': chart_query['traffic'],
                'colors': [
                    DEFAULT_COLORS[7],
                    DEFAULT_COLORS[0],
                    DEFAULT_COLORS[1],
                ],
            }
        },
    },
    'general_traffic': {
        'label': _('General Traffic'),
        'name': 'General Traffic',
        'key': 'traffic',
        'field_name': 'rx_bytes',
        'related_fields': ['tx_bytes'],
        'charts': {
            'general_traffic': {
                'type': 'stackedbar+lines',
                'calculate_total': True,
                'fill': 'none',
                'trace_type': {
                    'download': 'stackedbar',
                    'upload': 'stackedbar',
                    'total': 'lines',
                },
                'trace_order': ['total', 'download', 'upload'],
                'title': _('General Traffic'),
                'label': _('General Traffic'),
                'description': _(
                    'Network traffic of the whole network (total, download, upload).'
                ),
                'summary_labels': [
                    _('Total traffic'),
                    _('Total download traffic'),
                    _('Total upload traffic'),
                ],
                'unit': 'adaptive_prefix+B',
                'order': 240,
                'query': chart_query['general_traffic'],
                'query_default_param': {
                    'organization_id': '',
                    'ifname': '',
                    'location_id': '',
                    'floorplan_id': '',
                },
                'colors': [
                    DEFAULT_COLORS[7],
                    DEFAULT_COLORS[0],
                    DEFAULT_COLORS[1],
                ],
            }
        },
    },
    'clients': {
        'label': _('Clients'),
        'name': '{name}',
        'key': '{key}',
        'field_name': 'clients',
        'charts': {
            'wifi_clients': {
                'type': 'bar',
                'label': _('WiFi clients'),
                'title': _('WiFi clients: {ifname}'),
                'description': _(
                    'WiFi clients associated to the wireless interface "{ifname}".'
                ),
                'summary_labels': [_('Total Unique WiFi clients')],
                'unit': '',
                'order': 230,
                'query': chart_query['wifi_clients'],
            }
        },
    },
    'general_clients': {
        'label': _('General WiFi Clients'),
        'name': _('General WiFi Clients'),
        'key': 'wifi_clients',
        'field_name': 'clients',
        'charts': {
            'gen_wifi_clients': {
                'type': 'bar',
                'label': _('General WiFi Clients'),
                'title': _('General WiFi Clients'),
                'description': _('Unique WiFi clients count of the entire network.'),
                'summary_labels': [_('Total Unique WiFi clients')],
                'unit': '',
                'order': 230,
                'query': chart_query['general_wifi_clients'],
                'query_default_param': {
                    'organization_id': '',
                    'location_id': '',
                    'floorplan_id': '',
                },
            }
        },
    },
    'disk': {
        'label': _('Disk usage'),
        'name': 'Disk usage',
        'key': 'disk',
        'field_name': 'used_disk',
        'charts': {
            'disk': {
                'type': 'scatter',
                'title': _('Disk Usage'),
                'description': _(
                    'Disk usage in percentage, calculated using all the available partitions.'
                ),
                'summary_labels': [_('Disk Usage')],
                'unit': '%',
                'colors': [DEFAULT_COLORS[9]],
                'order': 270,
                'query': chart_query['disk'],
            }
        },
        'alert_settings': {'operator': '>', 'threshold': 90, 'tolerance': 0},
        'notification': {
            'problem': {
                'verbose_name': 'Disk usage PROBLEM',
                'verb': _('is experiencing a peak in'),
                'level': 'warning',
                'email_subject': _(
                    '[{site.name}] PROBLEM: {notification.target} {notification.verb} disk usage'
                ),
                'message': _(
                    'The device [{notification.target}]({notification.target_link}) '
                    '{notification.verb} disk usage which has gone over '
                    '{notification.actor.alertsettings.threshold}%.'
                ),
            },
            'recovery': {
                'verbose_name': 'Disk usage RECOVERY',
                'verb': _('has returned to normal levels'),
                'level': 'info',
                'email_subject': _(
                    '[{site.name}] RECOVERY: {notification.target} disk usage '
                    '{notification.verb}'
                ),
                'message': (
                    'The device [{notification.target}]({notification.target_link}) '
                    'disk usage {notification.verb}.'
                ),
            },
        },
    },
    'memory': {
        'label': _('Memory usage'),
        'name': 'Memory usage',
        'key': 'memory',
        'field_name': 'percent_used',
        'related_fields': [
            'total_memory',
            'free_memory',
            'buffered_memory',
            'shared_memory',
            'cached_memory',
            'available_memory',
        ],
        'charts': {
            'memory': {
                'type': 'scatter',
                'title': _('Memory Usage'),
                'description': _('Percentage of memory (RAM) being used.'),
                'summary_labels': [_('Memory Usage')],
                'unit': '%',
                'colors': [DEFAULT_COLORS[4]],
                'order': 250,
                'query': chart_query['memory'],
            }
        },
        'alert_settings': {'operator': '>', 'threshold': 95, 'tolerance': 5},
        'notification': {
            'problem': {
                'verbose_name': 'Memory usage PROBLEM',
                'verb': _('is experiencing a peak in'),
                'level': 'warning',
                'email_subject': _(
                    '[{site.name}] PROBLEM: {notification.target} {notification.verb} RAM usage'
                ),
                'message': _(
                    'The device [{notification.target}]({notification.target_link}) '
                    '{notification.verb} RAM usage which has gone '
                    'over {notification.actor.alertsettings.threshold}%.'
                ),
            },
            'recovery': {
                'verbose_name': 'Memory usage RECOVERY',
                'verb': _('has returned to normal levels'),
                'level': 'info',
                'email_subject': _(
                    '[{site.name}] RECOVERY: {notification.target} RAM usage {notification.verb}'
                ),
                'message': (
                    'The device [{notification.target}]({notification.target_link}) RAM usage '
                    '{notification.verb}.'
                ),
            },
        },
    },
    'cpu': {
        'label': _('CPU usage'),
        'name': 'CPU usage',
        'key': 'cpu',
        'field_name': 'cpu_usage',
        'related_fields': ['load_1', 'load_5', 'load_15'],
        'charts': {
            'cpu': {
                'type': 'scatter',
                'title': _('CPU Load'),
                'description': _(
                    'Average CPU load, measured using the Linux load averages, '
                    'taking into account the number of available CPUs.'
                ),
                'summary_labels': [_('CPU Load')],
                'unit': '%',
                'colors': [DEFAULT_COLORS[-3]],
                'order': 260,
                'query': chart_query['cpu'],
            }
        },
        'alert_settings': {'operator': '>', 'threshold': 90, 'tolerance': 5},
        'notification': {
            'problem': {
                'verbose_name': 'CPU usage PROBLEM',
                'verb': _('is experiencing a peak in'),
                'level': 'warning',
                'email_subject': _(
                    '[{site.name}] PROBLEM: {notification.target} {notification.verb} CPU usage'
                ),
                'message': _(
                    'The device [{notification.target}]({notification.target_link}) '
                    '{notification.verb} CPU usage which has gone '
                    'over {notification.actor.alertsettings.threshold}%.'
                ),
            },
            'recovery': {
                'verbose_name': 'CPU usage RECOVERY',
                'verb': _('has returned to normal levels'),
                'level': 'info',
                'email_subject': _(
                    '[{site.name}] RECOVERY: {notification.target} CPU usage {notification.verb}'
                ),
                'message': (
                    'The device [{notification.target}]({notification.target_link}) '
                    'CPU usage {notification.verb}.'
                ),
            },
        },
    },
    # modem-manager charts
    'signal_strength': {
        'label': _('Signal Strength'),
        'name': '{name}',
        'key': '{key}',
        'field_name': 'signal_strength',
        'related_fields': ['signal_power'],
        'charts': {
            'signal_strength': {
                'type': 'scatter',
                'fill': 'none',
                'yaxis': {'zeroline': False},
                'title': _('Signal Strength'),
                'colors': (DEFAULT_COLORS[3], DEFAULT_COLORS[0]),
                'description': _('Signal Strength and Signal Power, measured in dBm.'),
                'summary_labels': [
                    _('Average Signal Power'),
                    _('Average Signal Strength'),
                ],
                'unit': _(' dBm'),
                'order': 205,
                'query': chart_query['signal_strength'],
            }
        },
    },
    'signal_quality': {
        'label': _('Signal Quality'),
        'name': '{name}',
        'key': '{key}',
        'field_name': 'signal_quality',
        'related_fields': ['snr'],
        'charts': {
            'signal_quality': {
                'type': 'scatter',
                'fill': 'none',
                'yaxis': {'zeroline': False},
                'title': _('Signal Quality'),
                'colors': (DEFAULT_COLORS[3], DEFAULT_COLORS[0]),
                'description': _(
                    _('Signal Quality and Signal to Noise Ratio (SNR), measured in dB.')
                ),
                'summary_labels': [
                    _('Average Signal Quality'),
                    _('Average Signal to Noise Ratio'),
                ],
                'unit': _(' dB'),
                'order': 206,
                'query': chart_query['signal_quality'],
            }
        },
    },
    'access_tech': {
        'label': _('Access Technology'),
        'name': '{name}',
        'key': '{key}',
        'field_name': 'access_tech',
        'charts': {
            'access_tech': {
                'type': 'bar',
                'title': _('Access Technology'),
                'description': _(
                    _(
                        'Shows the access technology (LTE, UTMS, CDMA1x, etc.) '
                        'in use on the device at the specified time.'
                    )
                ),
                'order': 207,
                'unit': '',
                'colorscale': {
                    'label': '',
                    'map': _get_access_tech(),
                    'max': 5,
                    'min': 0,
                    'scale': (
                        (0, '#b42a0c'),
                        (0.1, '#dd5817'),
                        (0.2, '#df7514'),
                        (0.5, '#efdd50'),
                        (0.8, '#67c368'),
                        (1, '#377873'),
                    ),
                    'fixed_value': 100,
                },
                'query': chart_query['access_tech'],
            }
        },
    },
    'iperf3': {
        'label': _('Iperf3'),
        'name': 'Iperf3',
        'key': 'iperf3',
        'field_name': 'iperf3_result',
        'related_fields': [
            'sent_bps_tcp',
            'received_bps_tcp',
            'sent_bytes_tcp',
            'received_bytes_tcp',
            'retransmits',
            'sent_bytes_udp',
            'sent_bps_udp',
            'jitter',
            'total_packets',
            'lost_packets',
            'lost_percent',
        ],
        'charts': {
            'bandwidth': {
                'type': 'scatter',
                'connect_points': True,
                'title': _('Bandwidth'),
                'fill': 'none',
                'description': _('Bitrate during Iperf3 test.'),
                'summary_labels': [
                    _('TCP bitrate'),
                    _('UDP bitrate'),
                ],
                'unit': 'adaptive_prefix+bps',
                'order': 280,
                'query': chart_query['bandwidth'],
                'colors': [
                    DEFAULT_COLORS[0],
                    DEFAULT_COLORS[3],
                ],
            },
            'transfer': {
                'type': 'scatter',
                'connect_points': True,
                'fill': 'none',
                'title': _('Transferred Data'),
                'description': _('Transferred Data during Iperf3 test.'),
                'summary_labels': [
                    _('TCP transferred data'),
                    _('UDP transferred data'),
                ],
                'unit': 'adaptive_prefix+B',
                'order': 290,
                'query': chart_query['transfer'],
                'colors': [
                    DEFAULT_COLORS[0],
                    DEFAULT_COLORS[3],
                ],
            },
            'retransmits': {
                'type': 'scatter',
                'connect_points': True,
                'title': _('Retransmits'),
                'description': _('Retransmits during Iperf3 test in TCP mode.'),
                'summary_labels': [_('Restransmits')],
                'unit': '',
                'order': 300,
                'query': chart_query['retransmits'],
                'colors': [DEFAULT_COLORS[-3]],
            },
            'jitter': {
                'type': 'scatter',
                'connect_points': True,
                'title': _('Jitter'),
                'description': _(
                    'Jitter is a variance in latency measured using Iperf3 utility in UDP mode.'
                ),
                'summary_labels': [
                    _('Jitter'),
                ],
                'unit': _(' ms'),
                'order': 330,
                'query': chart_query['jitter'],
                'colors': [DEFAULT_COLORS[4]],
            },
            'datagram': {
                'type': 'scatter',
                'fill': 'none',
                'connect_points': True,
                'title': _('Datagram'),
                'description': _(
                    '(Lost / Total) datagrams measured by Iperf3 test in UDP mode.'
                ),
                'summary_labels': [
                    _('Lost datagram'),
                    _('Total datagram'),
                ],
                'unit': '',
                'order': 340,
                'query': chart_query['datagram'],
                'colors': [DEFAULT_COLORS[3], DEFAULT_COLORS[2]],
            },
            'datagram_loss': {
                'type': 'scatter',
                'connect_points': True,
                'title': _('Datagram Loss'),
                'description': _(
                    'Indicates datagram loss % during Iperf3 test in UDP mode.'
                ),
                'summary_labels': [
                    _('Datagram loss'),
                ],
                'unit': '%',
                'order': 350,
                'query': chart_query['datagram_loss'],
                'colors': [DEFAULT_COLORS[3]],
            },
        },
    },
}

DEFAULT_CHARTS = {}

DEFAULT_DASHBOARD_TRAFFIC_CHART = {'__all__': ['wan', 'eth1', 'eth0.2']}


def _validate_metric_configuration(metric_config):
    assert 'label' in metric_config
    assert 'name' in metric_config
    assert 'key' in metric_config
    assert 'field_name' in metric_config
    if 'alert_field' in metric_config:
        # ensure only valid alert_field is present
        alert_fields = [metric_config['field_name']] + metric_config['related_fields']
        assert metric_config['alert_field'] in alert_fields


def _validate_chart_configuration(chart_config):
    assert 'type' in chart_config
    assert 'title' in chart_config
    assert 'description' in chart_config
    assert 'order' in chart_config
    assert 'query' in chart_config
    if chart_config['query'] is None:
        assert 'unit' in chart_config
    if 'colorscale' in chart_config:
        assert 'max' in chart_config['colorscale']
        assert 'min' in chart_config['colorscale']
        assert 'label' in chart_config['colorscale']
        assert 'scale' in chart_config['colorscale']


def register_metric_notifications(metric_name, metric_config):
    if 'notification' not in metric_config:
        return
    register_notification_type(
        f'{metric_name}_problem', metric_config['notification']['problem']
    )
    register_notification_type(
        f'{metric_name}_recovery', metric_config['notification']['recovery']
    )


def unregister_metric_notifications(metric_name):
    metric_config = DEFAULT_METRICS[metric_name]
    if 'notification' not in metric_config:
        return
    unregister_notification_type(f'{metric_name}_problem')
    unregister_notification_type(f'{metric_name}_recovery')


def get_metric_configuration():
    additional_metrics = deepcopy(app_settings.ADDITIONAL_METRICS)
    for metric_name in list(additional_metrics.keys()):
        if additional_metrics[metric_name].get('partial', False):
            # A partial configuration can be defined in the settings.py
            # with OPENWISP_MONITORING_METRICS setting to override
            # metrics that are added with register_metric method in
            # other django apps.
            # Since, the partial configuration could be defined to
            # override limited fields in the configuration, hence
            # we don't validate the configuration here. Instead.
            # configuration is validated in the register_metric method.
            del additional_metrics[metric_name]
    metrics = deep_merge_dicts(DEFAULT_METRICS, additional_metrics)
    # ensure configuration is not broken
    for metric_config in metrics.values():
        _validate_metric_configuration(metric_config)
    return metrics


def get_metric_configuration_choices():
    metrics = get_metric_configuration()
    choices = []
    for key in sorted(metrics.keys()):
        label = metrics[key]['label']
        choices.append((key, label))
    return choices


def register_metric(metric_name, metric_config):
    """Registers a new metric configuration."""
    if not isinstance(metric_name, str):
        raise ImproperlyConfigured('Metric configuration name should be type "str".')
    if not isinstance(metric_config, dict):
        raise ImproperlyConfigured('Metric configuration should be type "dict".')
    if metric_name in DEFAULT_METRICS:
        raise ImproperlyConfigured(
            f'{metric_name} is an already registered Metric Configuration.'
        )
    if metric_name in app_settings.ADDITIONAL_METRICS:
        # There is partial configuration present for this "metric_name" in
        # ADDITIONAL_METRICS. We need to merge the partial configuration with
        # the registered metric before validating. Otherwise, users won't be
        # able to override registered metrics using OPENWISP_MONITORING_METRICS
        # setting.
        metric_config = deep_merge_dicts(
            metric_config,
            app_settings.ADDITIONAL_METRICS[metric_name],
        )
        metric_config.pop('partial', None)
    _validate_metric_configuration(metric_config)
    for chart in metric_config.get('charts', {}).values():
        _validate_chart_configuration(chart_config=chart)
    DEFAULT_METRICS.update({metric_name: metric_config})
    _register_metric_configuration_choice(metric_name, metric_config)
    register_metric_notifications(metric_name, metric_config)


def unregister_metric(metric_name):
    if not isinstance(metric_name, str):
        raise ImproperlyConfigured('Metric configuration name should be type "str".')
    if metric_name not in DEFAULT_METRICS:
        raise ImproperlyConfigured(f'No such Chart configuation "{metric_name}".')
    unregister_metric_notifications(metric_name)
    DEFAULT_METRICS.pop(metric_name)
    _unregister_metric_configuration_choice(metric_name)


def _register_metric_configuration_choice(metric_name, metric_config):
    name = metric_config.get('label', metric_name)
    METRIC_CONFIGURATION_CHOICES.append((metric_name, name))


def _unregister_metric_configuration_choice(metric_name):
    for index, (key, name) in enumerate(METRIC_CONFIGURATION_CHOICES):
        if key == metric_name:
            METRIC_CONFIGURATION_CHOICES.pop(index)
            return


def get_chart_configuration():
    metrics = get_metric_configuration()
    for metric in metrics.values():
        if 'charts' in metric:
            DEFAULT_CHARTS.update(metric['charts'])
    charts = deep_merge_dicts(DEFAULT_CHARTS, app_settings.ADDITIONAL_CHARTS)
    # ensure configuration is not broken
    for key, options in charts.items():
        _validate_chart_configuration(options)
    return charts


def get_chart_configuration_choices():
    charts = get_chart_configuration()
    choices = []
    for key in sorted(charts.keys()):
        label = charts[key].get('label', charts[key]['title'])
        choices.append((key, label))
    return choices


def register_chart(chart_name, chart_config):
    """Registers a new chart configuration."""
    if not isinstance(chart_name, str):
        raise ImproperlyConfigured('Chart name should be type "str".')
    if not isinstance(chart_config, dict):
        raise ImproperlyConfigured('Chart configuration should be type "dict".')
    if chart_name in get_chart_configuration():
        raise ImproperlyConfigured(
            f'{chart_name} is an already registered Chart Configuration.'
        )
    _validate_chart_configuration(chart_config)
    DEFAULT_CHARTS.update({chart_name: chart_config})
    _register_chart_configuration_choice(chart_name, chart_config)


def unregister_chart(chart_name):
    if not isinstance(chart_name, str):
        raise ImproperlyConfigured('Chart configuration name should be type "str"')
    if chart_name not in DEFAULT_CHARTS:
        raise ImproperlyConfigured(f'No such Chart configuation "{chart_name}"')
    DEFAULT_CHARTS.pop(chart_name)
    _unregister_chart_configuration_choice(chart_name)


def _register_chart_configuration_choice(chart_name, chart_config):
    name = chart_config.get('label', chart_name)
    CHART_CONFIGURATION_CHOICES.append((chart_name, name))


def _unregister_chart_configuration_choice(chart_name):
    for index, (key, name) in enumerate(CHART_CONFIGURATION_CHOICES):
        if key == chart_name:
            CHART_CONFIGURATION_CHOICES.pop(index)
            return


def _clean_dashboard_traffic_chart():
    for interfaces in DEFAULT_DASHBOARD_TRAFFIC_CHART.values():
        for index in range(len(interfaces)):
            interfaces[index] = clean_timeseries_data_key(interfaces[index])


DEFAULT_DASHBOARD_TRAFFIC_CHART.update(app_settings.ADDITIONAL_DASHBOARD_TRAFFIC_CHART)
_clean_dashboard_traffic_chart()
METRIC_CONFIGURATION_CHOICES = get_metric_configuration_choices()
CHART_CONFIGURATION_CHOICES = get_chart_configuration_choices()



from django.urls import include, path

urlpatterns = [
    path('', include('openwisp_monitoring.device.api.urls', namespace='monitoring')),
    # The following endpoint was developed after "openwisp_monitoring.device.api"
    # which already used the "monitoring" namespace. The "monitoring_general"
    # namespace is used below to avoid changing the old naming scheme and
    # maintain backward compatibility.
    path(
        '',
        include(
            'openwisp_monitoring.monitoring.api.urls', namespace='monitoring_general'
        ),
    ),
]



import logging
import time

import requests
from django.core.exceptions import ObjectDoesNotExist
from influxdb.exceptions import InfluxDBServerError
from swapper import load_model

from openwisp_monitoring.db.backends import timeseries_db
from openwisp_monitoring.db.exceptions import TimeseriesWriteException

SELECT_QUERY_LIMIT = 1000
WRITE_BATCH_SIZE = 1000
READ_QUERY = (
    "SELECT {fields} FROM \"{measurement}\""
    " WHERE content_type='{content_type_key}'"
    " AND object_id='{object_id}'"
)
DELETE_QUERY = (
    "DROP SERIES FROM \"{old_measurement}\""
    " WHERE content_type='{content_type_key}'"
    " AND object_id='{object_id}'"
)
CHUNK_SIZE = 1000
EXCLUDED_MEASUREMENTS = [
    'ping',
    'config_applied',
    'clients',
    'disk',
    'memory',
    'cpu',
    'signal_strength',
    'signal_quality',
    'access_tech',
    'device_data',
    'traffic',
    'wifi_clients',
]

logger = logging.getLogger(__name__)


def get_writable_data(read_data, tags, old_measurement, new_measurement):
    """Prepares data that can be written by "timeseries_db.db.write_points"."""
    write_data = []
    for data_point in read_data.get_points(measurement=old_measurement):
        data = {
            'time': data_point.pop('time'),
            'measurement': new_measurement,
            'tags': tags,
        }
        data['fields'] = data_point
        write_data.append(data)
    return write_data, len(write_data)


def retry_until_success(func, *args, **kwargs):
    sleep_time = 1
    while True:
        try:
            return func(*args, **kwargs)
        except (
            requests.exceptions.ConnectionError,
            TimeseriesWriteException,
            InfluxDBServerError,
            timeseries_db.client_error,
        ) as error:
            if 'points beyond retention policy dropped' in str(error):
                # When writing data to the InfluxDB, if the measurement
                # points are older than the retention policy the
                # InfluxDB returns a HTTP 400 response. Retrying this
                # operation will again result in HTTP 400, hence
                # this error is assumed as success.
                return True
            sleep_time *= 2
            time.sleep(sleep_time)
            logger.warning(
                (
                    f'Error encountered while executing {func.__name__}'
                    f' with args: {args} and kwargs: {kwargs}.\n'
                    f'Retrying after {sleep_time} seconds.\n'
                    f'Error: {error}'
                )
            )
            continue


def migrate_influxdb_data(
    configuration,
    new_measurement,
    read_query=READ_QUERY,
    delete_query=None,
):
    Metric = load_model('monitoring', 'Metric')
    metric_qs = Metric.objects.filter(configuration=configuration, key=new_measurement)
    for metric in metric_qs.iterator(chunk_size=CHUNK_SIZE):
        old_measurement = metric.main_tags.get('ifname')
        fields = ','.join(['time', metric.field_name, *metric.related_fields])
        query = (f"{read_query} ORDER BY time DESC LIMIT {SELECT_QUERY_LIMIT}").format(
            fields=fields,
            measurement=old_measurement,
            content_type_key=metric.content_type_key,
            object_id=metric.object_id,
        )
        # Read and write data in batches to avoid loading
        # all data in memory at once.
        offset = 0
        migrated_rows = 0
        read_data = retry_until_success(timeseries_db.query, query, epoch='s')
        while read_data:
            write_data, write_data_count = get_writable_data(
                read_data, metric.tags, old_measurement, new_measurement=new_measurement
            )
            start = offset
            end = offset + min(write_data_count, SELECT_QUERY_LIMIT)
            response = retry_until_success(
                timeseries_db._write,
                write_data,
                timeseries_db.db_name,
                retention_policy=None,
            )
            if response is True:
                logger.info(
                    f'Successfully written rows for "{metric} (id:{metric.id})"'
                    f' from {start} to {end}.'
                )
                migrated_rows += write_data_count
                offset += SELECT_QUERY_LIMIT
            else:
                logger.warning(
                    f'Error encountered in writing data for "{metric} (id:{metric.id})"'
                    f' from {start} to {end}: {response["error"]}. It will be retried.'
                )
            read_data = retry_until_success(
                timeseries_db.query, f'{query} OFFSET {offset}', epoch='s'
            )
        logger.info(f'Migrated {migrated_rows} row(s) for "{metric} (id:{metric.id})".')

        # Delete data that has been migrated
        if delete_query:
            retry_until_success(
                timeseries_db.query,
                delete_query.format(
                    old_measurement=old_measurement,
                    content_type_key=metric.content_type_key,
                    object_id=metric.object_id,
                ),
            )
            logger.info(f'Deleted old measurements for "{metric} (id:{metric.id})".')


def migrate_wifi_clients():
    read_query = f"{READ_QUERY} AND clients != ''"
    migrate_influxdb_data(
        new_measurement='wifi_clients',
        configuration='clients',
        read_query=read_query,
    )
    logger.info('"wifi_clients" measurements successfully migrated.')


def migrate_traffic_data():
    migrate_influxdb_data(new_measurement='traffic', configuration='traffic')
    logger.info('"traffic" measurements successfully migrated.')


def migrate_signal_strength_data():
    migrate_influxdb_data(new_measurement='signal', configuration='signal_strength')
    logger.info('"signal_strength" measurements successfully migrated.')


def migrate_signal_quality_data():
    migrate_influxdb_data(new_measurement='signal', configuration='signal_quality')
    logger.info('"signal_quality" measurements successfully migrated.')


def migrate_access_tech_data():
    migrate_influxdb_data(
        new_measurement='signal',
        configuration='access_tech',
        delete_query=DELETE_QUERY,
    )
    logger.info('"access_tech" measurements successfully migrated.')


def requires_migration():
    """Indicates whether influxdb data migration is necessary.

    Returns "False" if all measurements presents in InfluxDB are present
    in EXCLUDED_MEASUREMENTS. This means that there are no interface
    specific measurements. Otherwise, returns "True".
    """
    tsdb_measurements = timeseries_db.db.get_list_measurements()
    for measurement in tsdb_measurements:
        if measurement['name'] not in EXCLUDED_MEASUREMENTS:
            return True
    return False


def migrate_influxdb_structure():
    if not requires_migration():
        logger.info(
            'Timeseries data migration is already migrated. Skipping migration!'
        )
        return
    # IMPORTANT: Do not change order of below functions.
    # Lookup using fields not supported in WHERE clause during deletion.
    # "migrate_access_tech_data" at the end because it is deletes the metrics.
    migrate_wifi_clients()
    migrate_traffic_data()
    migrate_signal_strength_data()
    migrate_signal_quality_data()
    migrate_access_tech_data()
    logger.info('Timeseries data migration completed.')


def _forward_migrate_metric(metric_model, configuration, new_key, add_org_tag=True):
    metric_qs = metric_model.objects.filter(configuration=configuration).exclude(
        key__in=EXCLUDED_MEASUREMENTS
    )
    updated_metrics = []
    for metric in metric_qs.iterator(chunk_size=CHUNK_SIZE):
        try:
            assert add_org_tag is True
            extra_tags = {'organization_id': str(metric.content_object.organization_id)}
        except (AssertionError, ObjectDoesNotExist):
            extra_tags = {}
        metric.main_tags = {
            'ifname': metric.key,
        }
        metric.extra_tags.update(extra_tags)
        metric.key = new_key
        updated_metrics.append(metric)
        if len(updated_metrics) > CHUNK_SIZE:
            metric_model.objects.bulk_update(
                updated_metrics, fields=['main_tags', 'extra_tags', 'key']
            )
            updated_metrics = []
    if updated_metrics:
        metric_model.objects.bulk_update(
            updated_metrics, fields=['main_tags', 'extra_tags', 'key']
        )


def update_metric_timeseries_structure_forward_migration(apps, schema_editor):
    """
    Updates metric objects to use static value for key and set
    interface name in the main tags
    """
    Metric = load_model('monitoring', 'Metric')
    _forward_migrate_metric(Metric, configuration='clients', new_key='wifi_clients')
    _forward_migrate_metric(Metric, configuration='traffic', new_key='traffic')
    _forward_migrate_metric(
        Metric, configuration='signal_strength', new_key='signal', add_org_tag=False
    )
    _forward_migrate_metric(
        Metric, configuration='signal_quality', new_key='signal', add_org_tag=False
    )
    _forward_migrate_metric(
        Metric, configuration='access_tech', new_key='signal', add_org_tag=False
    )


def update_metric_timeseries_structure_reverse_migration(
    apps, schema_editor, metric_keys=None
):
    """
    Reverse migration is required because of the
    the unique together condition implemented in
    Metric model.
    """
    Metric = load_model('monitoring', 'Metric')
    metric_keys = metric_keys or ['traffic', 'wifi_clients', 'signal']
    updated_metrics = []
    for metric in Metric.objects.filter(
        key__in=metric_keys,
    ).iterator(chunk_size=CHUNK_SIZE):
        metric.key = metric.main_tags['ifname']
        updated_metrics.append(metric)
        if len(updated_metrics) > CHUNK_SIZE:
            Metric.objects.bulk_update(updated_metrics, fields=['key'])
            updated_metrics = []
    if updated_metrics:
        Metric.objects.bulk_update(updated_metrics, fields=['key'])



import json
import logging

from celery import shared_task
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from swapper import load_model

from openwisp_utils.tasks import OpenwispCeleryTask

from .settings import CHECKS_LIST

logger = logging.getLogger(__name__)


def get_check_model():
    return load_model('check', 'Check')


@shared_task(time_limit=2 * 60 * 60)
def run_checks(checks=None):
    """Runs all the checks.

    Retrieves the id of all active checks in chunks of 2000 items and
    calls the ``perform_check`` task (defined below) for each of them.

    This allows to enqueue all the checks that need to be performed and
    execute them in parallel with multiple workers if needed.
    """
    # If checks is None, We should execute all the checks
    if checks is None:
        checks = CHECKS_LIST

    if not isinstance(checks, list):
        raise ImproperlyConfigured(
            f'Check path {checks} should be of type "list"'
        )  # pragma: no cover
    if not all(check_path in CHECKS_LIST for check_path in checks):
        raise ImproperlyConfigured(
            f'Check path {checks} should be in {CHECKS_LIST}'
        )  # pragma: no cover

    iterator = (
        get_check_model()
        .objects.filter(is_active=True, check_type__in=checks)
        .only('id')
        .values('id')
        .iterator()
    )
    for check in iterator:
        perform_check.delay(check['id'])


@shared_task(time_limit=30 * 60)
def perform_check(uuid):
    """Performs check with specified uuid.

    Retrieves check according to the passed UUID and calls the
    ``perform_check()`` method.
    """
    try:
        check = get_check_model().objects.get(pk=uuid)
    except ObjectDoesNotExist:
        logger.warning(f'The check with uuid {uuid} has been deleted')
        return
    result = check.perform_check()
    if settings.DEBUG:  # pragma: nocover
        print(json.dumps(result, indent=4, sort_keys=True))


@shared_task(base=OpenwispCeleryTask)
def auto_create_ping(
    model, app_label, object_id, check_model=None, content_type_model=None
):
    """Implements the auto creation of the ping check.

    Called by django signal (dispatch_uid: auto_ping) registered in check
    app's apps.py file.
    """
    Check = check_model or get_check_model()
    ping_path = 'openwisp_monitoring.check.classes.Ping'
    has_check = Check.objects.filter(
        object_id=object_id, content_type__model='device', check_type=ping_path
    ).exists()
    # create new check only if necessary
    if has_check:
        return
    content_type_model = content_type_model or ContentType
    ct = content_type_model.objects.get_by_natural_key(app_label=app_label, model=model)
    check = Check(
        name='Ping', check_type=ping_path, content_type=ct, object_id=object_id
    )
    check.full_clean()
    check.save()


@shared_task(base=OpenwispCeleryTask)
def auto_create_config_check(
    model, app_label, object_id, check_model=None, content_type_model=None
):
    """Implements the auto creation of the config modified check.

    Called by openwisp_monitoring.check.models.auto_config_check_receiver.
    """
    Check = check_model or get_check_model()
    config_check_path = 'openwisp_monitoring.check.classes.ConfigApplied'
    has_check = Check.objects.filter(
        object_id=object_id, content_type__model='device', check_type=config_check_path
    ).exists()
    # create new check only if necessary
    if has_check:
        return
    content_type_model = content_type_model or ContentType
    ct = content_type_model.objects.get_by_natural_key(app_label=app_label, model=model)
    check = Check(
        name='Configuration Applied',
        check_type=config_check_path,
        content_type=ct,
        object_id=object_id,
    )
    check.full_clean()
    check.save()


@shared_task(base=OpenwispCeleryTask)
def auto_create_iperf3_check(
    model, app_label, object_id, check_model=None, content_type_model=None
):
    """Implements the auto creation of the iperf3 check.

    Called by the
    openwisp_monitoring.check.models.auto_iperf3_check_receiver.
    """
    Check = check_model or get_check_model()
    iperf3_check_path = 'openwisp_monitoring.check.classes.Iperf3'
    has_check = Check.objects.filter(
        object_id=object_id, content_type__model='device', check_type=iperf3_check_path
    ).exists()
    # create new check only if necessary
    if has_check:
        return
    content_type_model = content_type_model or ContentType
    ct = content_type_model.objects.get_by_natural_key(app_label=app_label, model=model)
    check = Check(
        name='Iperf3',
        check_type=iperf3_check_path,
        content_type=ct,
        object_id=object_id,
    )
    check.full_clean()
    check.save()



import json
import logging
from collections import OrderedDict
from copy import deepcopy
from datetime import date, datetime, timedelta

from cache_memoize import cache_memoize
from dateutil.parser import parse as parse_date
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.validators import MaxValueValidator
from django.db import IntegrityError, models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from jsonfield import JSONField
from openwisp_notifications.signals import notify
from pytz import timezone as tz
from pytz import utc
from swapper import get_model_name

from openwisp_monitoring.monitoring.utils import clean_timeseries_data_key
from openwisp_utils.base import TimeStampedEditableModel

from ...db import default_chart_query, timeseries_db
from ...settings import CACHE_TIMEOUT, DEFAULT_CHART_TIME
from ..configuration import (
    CHART_CONFIGURATION_CHOICES,
    DEFAULT_COLORS,
    METRIC_CONFIGURATION_CHOICES,
    get_chart_configuration,
    get_metric_configuration,
)
from ..exceptions import InvalidChartConfigException, InvalidMetricConfigException
from ..signals import pre_metric_write, threshold_crossed
from ..tasks import _timeseries_batch_write, _timeseries_write, delete_timeseries

User = get_user_model()
logger = logging.getLogger(__name__)


def get_metric_cache_key(*args, **kwargs):
    cache_key = [
        'get_or_create_metric',
    ]
    cache_key.append(
        'key={}'.format(
            kwargs.get('key', kwargs.get('name', kwargs.get('configuration')))
        )
    )
    for key in ['content_type_id', 'object_id', 'configuration']:
        cache_key.append('{}={}'.format(key, kwargs.get(key)))
    cache_key.append('main_tags={}'.format(kwargs.get('main_tags', OrderedDict())))
    cache_key = ','.join(cache_key)
    return cache_key


class AbstractMetric(TimeStampedEditableModel):
    name = models.CharField(max_length=64)
    key = models.SlugField(
        max_length=64, blank=True, help_text=_('leave blank to determine automatically')
    )
    field_name = models.CharField(
        max_length=16,
        default='value',
        blank=True,
        help_text=_('leave blank to determine automatically'),
    )
    configuration = models.CharField(
        max_length=16, null=True, choices=METRIC_CONFIGURATION_CHOICES
    )
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.CharField(max_length=36, db_index=True, blank=True, null=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    main_tags = JSONField(
        _('main tags'),
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4},
        db_index=True,
    )
    extra_tags = JSONField(
        _('extra tags'),
        default=dict,
        blank=True,
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4},
    )
    # NULL means the health has yet to be assessed
    is_healthy = models.BooleanField(default=None, null=True, blank=True, db_index=True)
    # Like "is_healthy", but respects tolerance of alert settings
    is_healthy_tolerant = models.BooleanField(default=None, null=True, blank=True)

    class Meta:
        abstract = True
        unique_together = (
            'key',
            'field_name',
            'content_type',
            'object_id',
            'main_tags',
        )

    def __str__(self):
        obj = self.content_object
        if not obj:
            return self.name
        model_name = obj.__class__.__name__
        return '{0} ({1}: {2})'.format(self.name, model_name, obj)

    def __setattr__(self, attrname, value):
        if attrname in ['main_tags', 'extra_tags']:
            value = self._sort_dict(value)
        return super().__setattr__(attrname, value)

    def clean(self):
        if self.key:
            return
        elif self.config_dict['key'] != '{key}':
            self.key = self.config_dict['key']
        else:
            self.key = self.codename

    def validate_alert_fields(self):
        # When field_name is not provided while creating a metric
        # then use config_dict['field_name] as metric field_name
        if self.config_dict['field_name'] != '{field_name}':
            if self.field_name in ['', 'value']:
                self.field_name = self.config_dict['field_name']
                return
            # field_name must be one of the metric fields
            alert_fields = [self.config_dict['field_name']] + self.related_fields
            if self.field_name not in alert_fields:
                raise ValidationError(
                    f'"{self.field_name}" must be one of the following metric fields ie. {alert_fields}'
                )

    def full_clean(self, *args, **kwargs):
        # The name of the metric will be the same as the
        # configuration chosen by the user only when the
        # name field is empty (useful for AlertSettingsInline)
        if not self.name:
            self.name = self.get_configuration_display()
        # clean up key before field validation
        self.key = self._makekey(self.key)
        # validate metric field_name for alerts
        self.validate_alert_fields()
        return super().full_clean(*args, **kwargs)

    @classmethod
    def post_delete_receiver(cls, instance, *args, **kwargs):
        delete_timeseries.delay(instance.key, instance.tags)

    @classmethod
    def _get_or_create(
        cls,
        **kwargs,
    ):
        """Gets or creates a metric.

        Like ``get_or_create`` method of django model managers but with
        validation before creation.
        """
        if 'key' in kwargs:
            kwargs['key'] = cls._makekey(kwargs['key'])
        try:
            lookup_kwargs = deepcopy(kwargs)
            if lookup_kwargs.get('name'):
                del lookup_kwargs['name']
            extra_tags = lookup_kwargs.pop('extra_tags', {})
            metric = cls._get_metric(**lookup_kwargs)
            created = False
            if extra_tags != metric.extra_tags:
                metric.extra_tags.update(kwargs['extra_tags'])
                metric.extra_tags = cls._sort_dict(metric.extra_tags)
                metric.save()
        except cls.DoesNotExist:
            try:
                metric = cls(**kwargs)
                metric.full_clean()
                metric.save()
                created = True
            except IntegrityError:
                # Potential race conditions may arise when multiple
                # celery workers concurrently write data to InfluxDB.
                # These simultaneous writes can result in the database
                # processing transactions from another "metric.save()"
                # call, potentially leading to IntegrityError exceptions.
                return cls._get_or_create(**kwargs)
        return metric, created

    @classmethod
    @cache_memoize(CACHE_TIMEOUT, key_generator_callable=get_metric_cache_key)
    def _get_metric(cls, *args, **kwargs):
        return cls.objects.get(**kwargs)

    @classmethod
    def invalidate_cache(cls, instance, *args, **kwargs):
        if kwargs.get('created', False):
            return
        cls._get_metric.invalidate(
            **{
                'name': instance.name,
                'key': instance.key,
                'content_type_id': instance.content_type_id,
                'object_id': instance.object_id,
                'configuration': instance.configuration,
                'main_tags': instance.main_tags,
            }
        )

    @property
    def codename(self):
        """Identifier stored in timeseries db."""
        return self._makekey(self.name)

    @property
    def config_dict(self):
        try:
            return get_metric_configuration()[self.configuration]
        except KeyError:
            raise InvalidMetricConfigException(
                f'Invalid metric configuration: "{self.configuration}"'
            )

    @property
    def related_fields(self):
        return self.config_dict.get('related_fields', [])

    # TODO: This method needs to be refactored when adding the other db
    @staticmethod
    def _makekey(value):
        """Produces a valid InfluxDB key from ``value``.

        Takes ``value`` as input argument and returs a valid InfluxDB key.
        """
        return clean_timeseries_data_key(value)

    @property
    def tags(self):
        tags = {}
        if self.content_type_id and self.object_id:
            tags.update(
                {
                    'content_type': self.content_type_key,
                    'object_id': str(self.object_id),
                }
            )
        if self.main_tags:
            tags.update(self.main_tags)
        if self.extra_tags:
            tags.update(self.extra_tags)
        return tags

    @staticmethod
    def _sort_dict(dict_):
        """Ensures the order of the keys in the dict is predictable."""
        if not isinstance(dict_, OrderedDict):
            return OrderedDict(sorted(dict_.items()))
        return dict_

    @property
    def content_type_key(self):
        try:
            content_type = ContentType.objects.get_for_id(self.content_type_id)
            return '.'.join(content_type.natural_key())
        except AttributeError:
            return None

    @property
    def alert_field(self):
        if self.field_name != self.config_dict['field_name']:
            return self.field_name
        return self.config_dict.get('alert_field', self.field_name)

    @property
    def alert_on_related_field(self):
        return self.alert_field in self.related_fields

    def _get_time(self, time):
        """If time is a string, it converts it to a datetime."""
        if isinstance(time, str):
            return parse_date(time)
        return time

    def _set_is_healthy(self, alert_settings, value):
        """Sets the value of "is_healthy" field when necessary.

        Executes only if "value" crosses the threshold defined in
        "alert_settings". Returns "True" if "is_healthy" field is changed.
        Otherwise, returns "None".

        This method does not take into account the alert settings
        tolerance, which is done by "_set_is_healthy_tolerant" method.
        """
        crossed = alert_settings._value_crossed(value)
        if (not crossed and self.is_healthy) or (crossed and self.is_healthy is False):
            return
        # problem: not within threshold limit
        elif crossed and self.is_healthy in [True, None]:
            self.is_healthy = False
        # ok: returned within threshold limit
        elif not crossed and self.is_healthy in [False, None]:
            self.is_healthy = True
        return True

    def _set_is_healthy_tolerant(
        self, alert_settings, value, time, retention_policy, send_alert
    ):
        """Sets the value of "is_tolerance_healthy" if necessary.

        Executes only if "value" crosses the threshold for more than the
        amount of seconds defined in the alert_settings "tolerance" field.
        It also sends the notification if required. Returns "None" if
        value of "is_healthy_tolerant" is unchanged. Returns "True" if it
        is the first metric write within threshold. Returns "False" in
        other cases.

        This method is similar to "_set_is_healthy" but it takes into
        account the alert settings tolerance so it's slightly different
        and more complex.
        """
        time = self._get_time(time)
        crossed = alert_settings._is_crossed_by(value, time, retention_policy)
        first_time = False
        # situation has not changed
        if (not crossed and self.is_healthy_tolerant) or (
            crossed and self.is_healthy_tolerant is False
        ):
            return
        # problem: not within threshold limit
        elif crossed and self.is_healthy_tolerant in [True, None]:
            if self.is_healthy_tolerant is None:
                first_time = True
            self.is_healthy_tolerant = False
            notification_type = f'{self.configuration}_problem'
        # ok: returned within threshold limit
        elif not crossed and self.is_healthy_tolerant is False:
            self.is_healthy_tolerant = True
            notification_type = f'{self.configuration}_recovery'
        # First metric write within threshold
        elif not crossed and self.is_healthy_tolerant is None:
            self.is_healthy_tolerant = True
            first_time = True

        # If we got to this point, it means we have to send an alert,
        # because the metric has been crossed for more than the
        # tolerated amount of time. There's one exception though:
        # if the device is new, its status will be unknown and the metric
        # will become healthy for the first time, in this case we do not need
        # to send an alert.
        if (
            not (first_time and self.is_healthy_tolerant)
            and alert_settings.is_active
            and send_alert
        ):
            self._notify_users(notification_type, alert_settings)
        return first_time

    def _is_historical_data(self, time):
        """
        Data older than 5 minutes is considered historical data.
        """
        recent_time = timezone.now() - timedelta(minutes=5)
        return time < recent_time

    def check_threshold(self, value, time=None, retention_policy=None, send_alert=True):
        """Checks if the threshold is crossed and notifies users accordingly"""
        try:
            alert_settings = self.alertsettings
        except ObjectDoesNotExist:
            return

        is_healthy_changed = self._set_is_healthy(alert_settings, value)
        tolerance_healthy_changed_first_time = self._set_is_healthy_tolerant(
            alert_settings, value, time, retention_policy, send_alert
        )
        is_healthy_tolerant_changed = tolerance_healthy_changed_first_time is not None
        # Do nothing if none of the fields changed.
        if not is_healthy_changed and not is_healthy_tolerant_changed:
            return
        update_fields = []
        if is_healthy_changed:
            update_fields.append('is_healthy')
        if is_healthy_tolerant_changed:
            update_fields.append('is_healthy_tolerant')
        self.save(update_fields=update_fields)
        threshold_crossed.send(
            sender=self.__class__,
            alert_settings=alert_settings,
            metric=self,
            target=self.content_object,
            first_time=tolerance_healthy_changed_first_time,
            tolerance_crossed=is_healthy_tolerant_changed,
        )

    def write(
        self,
        value,
        current=False,
        time=None,
        database=None,
        check=True,
        extra_values=None,
        retention_policy=None,
        send_alert=True,
        write=True,
    ):
        """write timeseries data"""
        values = {self.field_name: value}
        if extra_values and isinstance(extra_values, dict):
            for key in extra_values.keys():
                if not self.related_fields or key not in self.related_fields:
                    raise ValueError(f'"{key}" not defined in metric configuration')
            values.update(extra_values)
        signal_kwargs = dict(
            sender=self.__class__,
            metric=self,
            values=values,
            time=time,
            current=current,
        )
        pre_metric_write.send(**signal_kwargs)
        timestamp = time or timezone.now()
        if isinstance(timestamp, str):
            timestamp = parse_date(timestamp)
        options = dict(
            tags=self.tags,
            timestamp=timestamp.isoformat(),
            database=database,
            retention_policy=retention_policy,
            current=current,
            metric=self,
        )
        # check can be disabled,
        # mostly for automated testing and debugging purposes
        if check and not self._is_historical_data(timestamp):
            options["check_threshold_kwargs"] = {
                "value": value,
                "time": time,
                "retention_policy": retention_policy,
                "send_alert": send_alert,
            }

            # if alert_on_related_field then check threshold
            # on the related_field instead of field_name
            if self.alert_on_related_field:
                if not extra_values:
                    raise ValueError(
                        'write() missing keyword argument: "extra_values" required for alert on related field'
                    )
                if self.alert_field not in extra_values.keys():
                    raise ValueError(
                        f'"{key}" is not defined for alert_field in metric configuration'
                    )
                options['check_threshold_kwargs'].update(
                    {'value': extra_values[self.alert_field]}
                )
        if write:
            _timeseries_write(name=self.key, values=values, **options)
        return {'name': self.key, 'values': values, **options}

    @classmethod
    def batch_write(cls, raw_data):
        error_dict = {}
        write_data = []
        for metric, kwargs in raw_data:
            try:
                write_data.append(metric.write(**kwargs, write=False))
            except ValueError as error:
                error_dict[metric.key] = str(error)
        _timeseries_batch_write(write_data)
        if error_dict:
            raise ValueError(error_dict)

    def read(self, **kwargs):
        """reads timeseries data"""
        return timeseries_db.read(
            key=self.key, fields=self.field_name, tags=self.tags, **kwargs
        )

    def _notify_users(self, notification_type, alert_settings):
        """creates notifications for users"""
        opts = dict(sender=self, type=notification_type, action_object=alert_settings)
        if self.content_object is not None:
            opts['target'] = self.content_object
        notify.send(**opts)


class AbstractChart(TimeStampedEditableModel):
    metric = models.ForeignKey(
        get_model_name('monitoring', 'Metric'), on_delete=models.CASCADE
    )
    configuration = models.CharField(
        max_length=16, null=True, choices=CHART_CONFIGURATION_CHOICES
    )
    GROUP_MAP = {'1d': '10m', '3d': '20m', '7d': '1h', '30d': '24h', '365d': '7d'}
    DEFAULT_TIME = DEFAULT_CHART_TIME

    class Meta:
        abstract = True

    def __str__(self):
        return str(self.label) or self.metric.name

    def clean(self):
        self._clean_query()

    def _clean_query(self):
        try:
            timeseries_db.validate_query(self.query)
            timeseries_db.query(self.get_query())
        except timeseries_db.client_error as e:
            raise ValidationError({'configuration': e}) from e
        except InvalidChartConfigException as e:
            raise ValidationError({'configuration': str(e)}) from e

    @property
    def config_dict(self):
        try:
            return get_chart_configuration()[self.configuration]
        except KeyError as e:
            raise InvalidChartConfigException(
                f'Invalid chart configuration: "{self.configuration}"'
            ) from e

    @property
    def type(self):
        return self.config_dict['type']

    @property
    def fill(self):
        return self.config_dict.get('fill')

    @property
    def xaxis(self):
        return self.config_dict.get('xaxis', {})

    @property
    def yaxis(self):
        return self.config_dict.get('yaxis', {})

    @property
    def label(self):
        return self.config_dict.get('label') or self.title

    @property
    def trace_type(self):
        return self.config_dict.get('trace_type', {})

    @property
    def trace_order(self):
        return self.config_dict.get('trace_order', [])

    @property
    def trace_labels(self):
        return self.config_dict.get('trace_labels', {})

    @property
    def calculate_total(self):
        return self.config_dict.get('calculate_total', False)

    @property
    def connect_points(self):
        return self.config_dict.get('connect_points', False)

    @property
    def description(self):
        return self.config_dict['description'].format(
            metric=self.metric, **self.metric.tags
        )

    @property
    def title(self):
        return self.config_dict['title']

    @property
    def summary_labels(self):
        return self.config_dict.get('summary_labels')

    @property
    def order(self):
        return self.config_dict['order']

    @property
    def colors(self):
        colors = self.config_dict.get('colors')
        if not colors and self.summary_labels:
            summary_length = len(self.summary_labels)
            return DEFAULT_COLORS[0:summary_length]
        return colors

    @property
    def colorscale(self):
        return self.config_dict.get('colorscale')

    @property
    def unit(self):
        return self.config_dict.get('unit')

    @property
    def query(self):
        query = self.config_dict['query']
        if query:
            return query[timeseries_db.backend_name]
        return self._default_query

    @property
    def summary_query(self):
        query = self.config_dict.get('summary_query', None)
        if query:
            return query[timeseries_db.backend_name]

    @property
    def top_fields(self):
        return self.config_dict.get('top_fields', None)

    @property
    def _default_query(self):
        q = default_chart_query[0]
        if self.metric.object_id:
            q += default_chart_query[1]
        return q

    @classmethod
    def _get_group_map(cls, time=None):
        """Returns group map.

        Returns the chart group map for the specified days, otherwise the
        default Chart.GROUP_MAP is returned.
        """
        if (
            not time
            or time in cls.GROUP_MAP.keys()
            or not isinstance(time, str)
            or time[-1] != 'd'
        ):
            return cls.GROUP_MAP
        group = '10m'
        days = int(time.split('d')[0])
        # Use copy of class variable to avoid unpredictable results
        custom_group_map = cls.GROUP_MAP.copy()
        # custom grouping between 1 to 2 days
        if days > 0 and days < 3:
            group = '10m'
        # custom grouping between 3 to 6 days (base 5)
        elif days >= 3 and days < 7:
            group = str(5 * round(((days / 3) * 20) / 5)) + 'm'
        # custom grouping between 8 to 27 days
        elif days > 7 and days < 28:
            group = str(round(days / 7)) + 'h'
        # custom grouping between 1 month to 7 month
        elif days >= 30 and days <= 210:
            group = '1d'
        # custom grouping between 7 month to 1 year
        elif days > 210 and days <= 365:
            group = '7d'
        custom_group_map.update({time: group})
        return custom_group_map

    def get_query(
        self,
        time=DEFAULT_TIME,
        summary=False,
        fields=None,
        query=None,
        timezone=settings.TIME_ZONE,
        start_date=None,
        end_date=None,
        additional_params=None,
    ):
        query = query or self.query
        if summary and self.summary_query:
            query = self.summary_query
        additional_params = additional_params or {}
        params = self._get_query_params(time, start_date, end_date)
        params.update(additional_params)
        params.update({'start_date': start_date, 'end_date': end_date})
        if not params.get('organization_id') and self.config_dict.get('__all__', False):
            params['organization_id'] = ['__all__']
        return timeseries_db.get_query(
            self.type,
            params,
            time,
            self._get_group_map(time),
            summary,
            fields,
            query,
            timezone,
        )

    def get_top_fields(self, number):
        """Returns the top fields.

        Returns list of top ``number`` of fields (highest sum) of a
        measurement in the specified time range (descending order).
        """
        q = self._default_query.replace('{field_name}', '{fields}')
        params = self._get_query_params(self.DEFAULT_TIME)
        return timeseries_db._get_top_fields(
            query=q,
            chart_type=self.type,
            group_map=self._get_group_map(params['days']),
            number=number,
            params=params,
            time=self.DEFAULT_TIME,
        )

    def _get_query_params(self, time, start_date=None, end_date=None):
        m = self.metric
        params = dict(
            field_name=m.field_name,
            key=m.key,
            time=self._get_time(time, start_date, end_date),
            days=time,
        )
        if m.object_id:
            params.update(
                {
                    'content_type': m.content_type_key,
                    'object_id': m.object_id,
                    **m.tags,
                }
            )
        for key, value in self.config_dict.get('query_default_param', {}).items():
            params.setdefault(key, value)
        return params

    @classmethod
    def _get_time(cls, time, start_date=None, end_date=None):
        if start_date and end_date:
            return start_date
        if not isinstance(time, str):
            return str(time)
        if time in cls._get_group_map(time).keys():
            days = int(time.strip('d'))
            now = timezone.now()
            if days > 3:
                now = date(now.year, now.month, now.day)
            if days == 7:
                # subtract one day because we want to include
                # the current day in the time range
                days -= 1
            time = str(now - timedelta(days=days))[0:19]
        return time

    def read(
        self,
        decimal_places=2,
        time=DEFAULT_TIME,
        x_axys=True,
        timezone=settings.TIME_ZONE,
        start_date=None,
        end_date=None,
        additional_query_kwargs=None,
    ):
        additional_query_kwargs = additional_query_kwargs or {}
        traces = {}
        if x_axys:
            x = []
        try:
            query_kwargs = dict(
                time=time, timezone=timezone, start_date=start_date, end_date=end_date
            )
            query_kwargs.update(additional_query_kwargs)
            if self.top_fields:
                fields = self.get_top_fields(self.top_fields)
                data_query = self.get_query(fields=fields, **query_kwargs)
                summary_query = self.get_query(
                    fields=fields, summary=True, **query_kwargs
                )
            else:
                data_query = self.get_query(**query_kwargs)
                summary_query = self.get_query(summary=True, **query_kwargs)
            points = timeseries_db.get_list_query(data_query)
            summary = timeseries_db.get_list_query(summary_query)
        except timeseries_db.client_error as e:
            logging.error(e, exc_info=True)
            raise e
        for point in points:
            for key, value in point.items():
                if key == 'time':
                    continue
                traces.setdefault(key, [])
                if decimal_places and isinstance(value, (int, float)):
                    value = self._round(value, decimal_places)
                traces[key].append(value)
            time = datetime.fromtimestamp(point['time'], tz=tz(timezone)).strftime(
                '%Y-%m-%d %H:%M'
            )
            if x_axys:
                x.append(time)
        # prepare result to be returned
        # (transform chart data so its order is not random)
        result = {'traces': sorted(traces.items())}
        if x_axys:
            result['x'] = x
        # add summary
        if len(summary) > 0:
            result['summary'] = {}
            for key, value in summary[0].items():
                if key == 'time':
                    continue
                if not timeseries_db.validate_query(self.query):
                    value = None
                elif value:
                    value = self._round(value, decimal_places)
                result['summary'][key] = value
        return result

    def json(self, time=DEFAULT_TIME, **kwargs):
        try:
            # unit needs to be passed for chart_inline
            data = self.read(time=time)
            data.update(
                {
                    'unit': self.unit,
                    'trace_type': self.trace_type,
                    'trace_order': self.trace_order,
                    'calculate_total': self.calculate_total,
                    'connect_points': self.connect_points,
                    'colors': self.colors,
                }
            )
            return json.dumps(data, **kwargs, default=str)
        except KeyError as e:
            logger.warning(f'Got KeyError in Chart.json method: {e}')

    @staticmethod
    def _round(value, decimal_places):
        """Rounds value when necessary."""
        control = 1.0 / 10**decimal_places
        if value < control:
            decimal_places += 2
        return round(value, decimal_places)


class AbstractAlertSettings(TimeStampedEditableModel):
    _MINUTES_MAX = 60 * 24 * 7  # 7 days
    _MINUTES_HELP = (
        'for how many minutes should the threshold value be crossed before '
        'an alert is sent? A value of zero means the alert is sent immediately'
    )
    _ALERTSETTINGS_OPERATORS = (('<', _('less than')), ('>', _('greater than')))
    is_active = models.BooleanField(
        _('Alerts enabled'),
        default=True,
        help_text=_(
            'whether alerts are enabled for this metric, uncheck to '
            'disable this alert for this object and all users'
        ),
    )
    metric = models.OneToOneField(
        get_model_name('monitoring', 'Metric'), on_delete=models.CASCADE
    )
    custom_operator = models.CharField(
        _('operator'),
        max_length=1,
        choices=_ALERTSETTINGS_OPERATORS,
        null=True,
        blank=True,
    )
    custom_threshold = models.FloatField(
        _('threshold value'), help_text=_('threshold value'), blank=True, null=True
    )
    custom_tolerance = models.PositiveIntegerField(
        _('threshold tolerance'),
        validators=[MaxValueValidator(_MINUTES_MAX)],
        help_text=_(_MINUTES_HELP),
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True
        verbose_name = _('Alert settings')
        verbose_name_plural = verbose_name
        permissions = (
            ('add_alertsettings_inline', 'Can add Alert settings inline'),
            ('change_alertsettings_inline', 'Can change Alert settings inline'),
            ('delete_alertsettings_inline', 'Can delete Alert settings inline'),
            ('view_alertsettings_inline', 'Can view Alert settings inline'),
        )

    @classmethod
    def invalidate_cache(cls, instance, *args, **kwargs):
        Metric = instance.metric._meta.model
        Metric.invalidate_cache(instance.metric)

    def full_clean(self, *args, **kwargs):
        if self.custom_threshold == self.config_dict['threshold']:
            self.custom_threshold = None
        if self.custom_tolerance == self.config_dict['tolerance']:
            self.custom_tolerance = None
        if self.custom_operator == self.config_dict['operator']:
            self.custom_operator = None
        return super().full_clean(*args, **kwargs)

    @property
    def config_dict(self):
        return self.metric.config_dict.get(
            'alert_settings', {'operator': '<', 'threshold': 1, 'tolerance': 0}
        )

    @property
    def threshold(self):
        if self.custom_threshold is None:
            return self.config_dict['threshold']
        return self.custom_threshold

    @property
    def tolerance(self):
        if self.custom_tolerance is None:
            return self.config_dict['tolerance']
        return self.custom_tolerance

    @property
    def operator(self):
        if self.custom_operator is None:
            return self.config_dict['operator']
        return self.custom_operator

    def _value_crossed(self, current_value):
        threshold_value = self.threshold
        method = '__gt__' if self.operator == '>' else '__lt__'
        if isinstance(current_value, int):
            current_value = float(current_value)
        return getattr(current_value, method)(threshold_value)

    def _time_crossed(self, time):
        threshold_time = timezone.now() - timedelta(minutes=self.tolerance)
        return time < threshold_time

    @property
    def _tolerance_search_range(self):
        """Returns an amount of minutes to search.

        Allows sufficient room for checking if the tolerance has been
        trepassed. Minimum 15 minutes, maximum self._MINUTES_MAX * 1.05.
        """
        minutes = self.tolerance * 2
        minutes = minutes if minutes > 15 else 15
        minutes = minutes if minutes <= self._MINUTES_MAX else self._MINUTES_MAX * 1.05
        return int(minutes)

    def _is_crossed_by(self, current_value, time=None, retention_policy=None):
        """Answers the following question:

        do current_value and time cross the threshold and trepass the
        tolerance?
        """
        value_crossed = self._value_crossed(current_value)
        if value_crossed is NotImplemented:
            raise ValueError('Supplied value type not suppported')
        # no tolerance specified, return immediately
        if self.tolerance == 0:
            return value_crossed
        # tolerance is set, we must go back in time
        # to ensure the threshold is trepassed for enough time
        # if alert field is supplied, retrieve such field when reading
        # so that we can let the system calculate the threshold on it
        extra_fields = []
        if self.metric.alert_on_related_field:
            extra_fields = [self.metric.alert_field]
        if time is None:
            # retrieves latest measurements, ordered by most recent first
            points = self.metric.read(
                since=f'{self._tolerance_search_range}m',
                limit=None,
                order='-time',
                retention_policy=retention_policy,
                extra_fields=extra_fields,
            )
            # store a list with the results
            results = [value_crossed]
            # loop on each measurement starting from the most recent
            for i, point in enumerate(points, 1):
                # skip the first point because it was just added before this
                # check started and its value coincides with ``current_value``
                if i <= 1:
                    continue
                utc_time = utc.localize(datetime.utcfromtimestamp(point['time']))
                # did this point cross the threshold? Append to result list
                results.append(self._value_crossed(point[self.metric.alert_field]))
                # tolerance is trepassed
                if self._time_crossed(utc_time):
                    # if the latest results are consistent, the metric being
                    # monitored is not flapping and we can confidently return
                    # wheter the value crosses the threshold or not
                    if len(set(results)) == 1:
                        return value_crossed
                    # otherwise, the results are flapping, the situation has not changed
                    # we will return a value that will not trigger changes
                    return not self.metric.is_healthy_tolerant
                # otherwise keep looking back
                continue
            # the search has not yielded any conclusion
            # return result based on the current value and time
            time = timezone.now()
        return self._time_crossed(time) and value_crossed
